"""
Dependency
----------

Special fields that allow us to make use of NuPlone's dependency engine also
for validating required fields.
A required field only needs to be validated against being empty if the
condition is met, that means, if the checkbox that it depends on is ticked.
"""

from euphorie.content.user import BaseValidator
from euphorie.htmllaundry.z3cform import HtmlText
from euphorie.htmllaundry.z3cform import IHtmlText
from plonetheme.nuplone.z3cform.directives import Dependency
from plonetheme.nuplone.z3cform.interfaces import INuPloneFormLayer
from z3c.form.browser.text import TextWidget
from z3c.form.browser.textarea import TextAreaWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import IForm
from z3c.form.interfaces import ITextAreaWidget
from z3c.form.interfaces import ITextWidget
from z3c.form.interfaces import IValidator
from z3c.form.widget import FieldWidget
from zope import schema
from zope.component import adapter
from zope.interface import implementer
from zope.interface import implementer_only
from zope.interface import Interface
from zope.schema.interfaces import IFromUnicode


class ConditionalField:
    """Marker class for conditional fields.

    FIXME: this class is used to register a custom validator adapter
    for the copnditional text input and textarea fields.

    We should avoid using this mixin class and
    use a proper common interface instead.
    """


# Interfaces
class IConditionalTextLine(IFromUnicode):
    """A Text line that is only shown under certain conditions."""


class IConditionalHtmlText(IHtmlText):
    """HTML Text field that is only shown under certain conditions."""


class IConditionalTextWidget(ITextWidget):
    """Text widget that is only shown under certain conditions."""


class IConditionalHtmlTextWidget(ITextAreaWidget):
    """Text area widget that is only shown under certain conditions."""


# Fields
@implementer(IConditionalTextLine)
class ConditionalTextLine(schema.TextLine, ConditionalField):
    """A Text line that is only shown under certain conditions."""


@implementer(IConditionalHtmlText)
class ConditionalHtmlText(HtmlText, ConditionalField):
    """HTML Text field that is only shown under certain conditions."""


# Widgets
@implementer_only(IConditionalTextWidget)
class ConditionalTextWidget(TextWidget):
    """Text widget that is only shown under certain conditions."""

    klass = "conditional-text-widget"


@implementer_only(IConditionalHtmlTextWidget)
class ConditionalHtmlTextWidget(TextAreaWidget):
    """Text area widget that is only shown under certain conditions."""

    klass = "conditional-textarea-widget"


# Field widget factories
@adapter(IConditionalTextLine, INuPloneFormLayer)
@implementer(IFieldWidget)
def ConditionalTextFieldWidget(field, request):
    return FieldWidget(field, ConditionalTextWidget(request))


@adapter(IConditionalHtmlText, INuPloneFormLayer)
@implementer(IFieldWidget)
def ConditionalHtmlTextFieldWidget(field, request):
    return FieldWidget(field, ConditionalHtmlText(request))


# Validator
@implementer(IValidator)
@adapter(Interface, Interface, IForm, ConditionalField, Interface)
class ConditionalFieldValidator(BaseValidator):
    def validate(self, value):
        """Only validate for required if the condition for the field is met.

        If the field that our widget depends on is not checked, set the
        flag for ignoring validation for required.
        """
        widget = self.widget
        dependencies = getattr(self.widget.field, "_dependencies", [])
        dependencies = [dep for dep in dependencies if isinstance(dep, Dependency)]
        if len(dependencies):
            widgets = widget.__parent__
            for dep in dependencies:
                # We only care about "on" type dependencies
                if dep.op != "on":
                    continue
                d_widget = widgets[dep.field]
                d_value = self.request.get(d_widget.name)
                # The magic happens here: the flag is set on self.widget, but
                # this is not persisted. Then the usual validation method
                # of z3c.form.validation is called, with the updated widget.
                self.widget.ignoreRequiredOnValidation = not d_value

        return super().validate(value)



"""
Upload
------

Form and browser view for importing a previously exported survey in XML format.

view: @@upload
"""

from ..risk import EnsureInterface
from ..risk import IFrenchEvaluation
from ..risk import IKinneyEvaluation
from ..risk import IRisk
from ..user import LoginField
from ..user import validLoginValue
from Acquisition import aq_inner
from euphorie.content import MessageFactory as _
from euphorie.content.behaviors.hide_from_training import IHideFromTraining
from euphorie.content.behaviors.toolcategory import IToolCategory
from euphorie.content.utils import IToolTypesInfo
from io import BytesIO
from markdownify import markdownify
from plone.autoform.form import AutoExtensibleForm
from plone.base.utils import safe_bytes
from plone.dexterity.utils import createContentInContainer
from plone.namedfile import field as filefield
from plone.namedfile.file import NamedBlobImage
from Products.statusmessages.interfaces import IStatusMessage
from z3c.form import button
from z3c.form import form
from z3c.form.interfaces import WidgetActionExecutionError
from zope import schema
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.interface import Interface
from zope.interface import Invalid

import lxml.etree
import lxml.objectify
import mimetypes
import random


try:
    from base64 import decodebytes
except ImportError:
    # PY27
    from base64 import decodestring as decodebytes

ProfileQuestionLocationFields = [
    "label_multiple_present",
    "label_single_occurance",
    "label_multiple_occurances",
]

NSMAP = {None: "http://xml.simplon.biz/euphorie/survey/1.0"}
XMLNS = "{%s}" % NSMAP[None]
COMMA_REPLACEMENT = "__COMMA__"


def attr_unicode(node, attr, default=None):
    value = str(node.attrib.get(attr, "")).strip()
    if not value:
        return default
    return value


def attr_string(node, tag, attr, default=None):
    value = default
    element = getattr(node, tag, None)
    if element is not None:
        value = element.attrib.get(attr)
    return value


def attr_vocabulary(node, tag, field, default=None):
    value = node.get(tag, "").strip() if node else None
    if not value:
        return default
    try:
        return field.vocabulary.getTermByToken(value).value
    except LookupError:
        return default


def attr_bool(node, tag, attr, default=False):
    value = default
    element = getattr(node, tag, None)
    if element is not None:
        value = element.get(attr)
    return value == "true"


def attr_int(node, tag, attr, default=None):
    value = default
    element = getattr(node, tag, None)
    if element is not None:
        value = int(element.get(attr))
    return value


def el_unicode(
    node, tag, default=None, is_etranslate_compatible=False, convert_to_markdown=False
):
    value = getattr(node, tag, None)
    if value is None:
        return default
    if is_etranslate_compatible:
        # We need to remove the outer XML element e.g. <description>
        wrapped_xml = lxml.etree.tostring(value, encoding="unicode", pretty_print=True)
        end_of_first_tag = wrapped_xml.find(">") + 1
        start_of_last_tag = wrapped_xml.rfind("<")
        unwrapped_html = wrapped_xml[end_of_first_tag:start_of_last_tag]
        if convert_to_markdown:
            return markdownify(unwrapped_html)
        else:
            return unwrapped_html
    else:
        return str(value)


def el_string(node, tag, default=None):
    value = getattr(node, tag, None)
    if value is None:
        return default
    return str(value.text)


def el_bool(node, tag, default=False):
    value = getattr(node, tag, None)
    if value is None:
        return default
    return value.text == "true"


class IImportSector(Interface):
    """The fields used for importing a :obj:`euphorie.content.sector`."""

    sector_title = schema.TextLine(
        title=_("label_sector_title", default="Title of sector."),
        description=_(
            "help_sector_title",
            default="If you do not specify a title it will be taken "
            "from the input file.",
        ),
        required=False,
    )

    sector_login = LoginField(
        title=_("label_login_name", default="Login name"),
        required=True,
        constraint=validLoginValue,
    )

    surveygroup_title = schema.TextLine(
        title=_("label_surveygroup_title", default="Title of imported OiRA Tool"),
        description=_(
            "help_upload_surveygroup_title",
            default="If you do not specify a title it will be taken " "from the input.",
        ),
        required=False,
    )

    survey_title = schema.TextLine(
        title=_("label_upload_survey_title", default="Name for OiRA Tool version"),
        default=_("Standard"),
        required=True,
    )

    file = filefield.NamedFile(
        title=_("label_upload_filename", default="XML file"), required=True
    )


class IImportSurvey(Interface):
    """The fields used for importing a :obj:`euphorie.content.survey`."""

    surveygroup_title = schema.TextLine(
        title=_("label_surveygroup_title", default="Title of imported OiRA Tool"),
        description=_(
            "help_upload_surveygroup_title",
            default="If you do not specify a title it will be taken " "from the input.",
        ),
        required=False,
    )

    survey_title = schema.TextLine(
        title=_("label_upload_survey_title", default="Name for OiRA Tool version"),
        default=_("OiRA Tool import"),
        required=True,
    )

    file = filefield.NamedFile(
        title=_("label_upload_filename", default="XML file"), required=True
    )

    is_etranslate_compatible = schema.Bool(
        title=_(
            "label_is_etranslate_compatible",
            default="Import XML translation from eTranslate",
        ),
        required=False,
        default=False,
    )


class SurveyImporter:
    """Import a survey version from an XML file and create a new survey group
    and survey.

    This assumes the current context is a sector.
    """

    is_etranslate_compatible = False

    def __init__(self, context):
        self.context = context

    def ImportImage(self, node):
        """Import a base64 encoded image from an XML node.

        :param node: lxml.objectified XML node of image element
        :rtype: (:py:class:`NamedImage`, unicode) tuple
        """
        filename = attr_unicode(node, "filename")
        contentType = node.get("content-type", None)
        if not filename:
            basename = "image%d.%%s" % random.randint(1, 2**16)
            if contentType and "/" in contentType:
                filename = basename % contentType.split("/")[1]
            else:
                filename = basename % "jpg"
        image = NamedBlobImage(
            data=decodebytes(safe_bytes(str(node.text))),
            contentType=contentType,
            filename=filename,
        )
        if image.contentType is None and image.filename:
            image.contentType = mimetypes.guess_type(image.filename)[0]
        return (image, attr_unicode(node, "caption"))

    def ImportSolution(self, node, risk):
        """
        Create a new :obj:`euphorie.content.solution` object for a
        :obj:`euphorie.content.risk` given the details for a Solution as an XML
        node.

        :returns: :obj:`euphorie.content.solution`.
        """
        solution = createContentInContainer(risk, "euphorie.solution")
        solution.description = str(node.description)
        action = el_unicode(
            node,
            "action",
            is_etranslate_compatible=self.is_etranslate_compatible,
            convert_to_markdown=True,
        )
        solution.action = action or node.description
        solution.action_plan = str(getattr(node, "action-plan", ""))
        solution.prevention_plan = el_unicode(node, "prevention-plan")
        solution.requirements = el_unicode(node, "requirements")
        solution.external_id = attr_unicode(node, "external-id")
        return solution

    def ImportRisk(self, node, module):
        """
        Create a new :obj:`euphorie.content.risk` object for a
        :obj:`euphorie.content.module` given the details for a Risk as an XML
        node.

        :returns: :obj:`euphorie.content.risk`.
        """
        risk = createContentInContainer(module, "euphorie.risk", title=str(node.title))
        EnsureInterface(risk)
        risk.type = node.get("type")
        risk.description = el_unicode(
            node, "description", is_etranslate_compatible=self.is_etranslate_compatible
        )
        risk.problem_description = el_unicode(
            node,
            "problem-description",
            is_etranslate_compatible=self.is_etranslate_compatible,
        )
        risk.legal_reference = el_unicode(node, "legal-reference")
        risk.show_notapplicable = el_bool(node, "show-not-applicable")
        risk.external_id = attr_unicode(node, "external-id")

        if risk.type == "risk":
            em = getattr(node, "evaluation-method", None)
            risk.evaluation_method = em.text if em else "estimated"
            if risk.evaluation_method == "calculated":
                evaluation_algorithm = risk.evaluation_algorithm()
                if evaluation_algorithm == "kinney":
                    risk.default_probability = attr_vocabulary(
                        em,
                        "default-probability",
                        IKinneyEvaluation["default_probability"],
                    )
                    risk.default_frequency = attr_vocabulary(
                        em, "default-frequency", IKinneyEvaluation["default_frequency"]
                    )
                    risk.default_effect = attr_vocabulary(
                        em, "default-effect", IKinneyEvaluation["default_effect"]
                    )
                elif evaluation_algorithm == "french":
                    risk.default_severity = attr_vocabulary(
                        em, "default-severity", IFrenchEvaluation["default_severity"]
                    )
                    risk.default_frequency = attr_vocabulary(
                        em, "default-frequency", IFrenchEvaluation["default_frequency"]
                    )
            else:
                risk.default_priority = attr_vocabulary(
                    em, "default-priority", IRisk["default_priority"]
                )

        for index, child in enumerate(node.iterchildren(tag=XMLNS + "image")):
            postfix = "" if not index else str(index + 1)
            (image, caption) = self.ImportImage(child)
            setattr(risk, "image" + postfix, image)
            setattr(risk, "caption" + postfix, caption)
            if index == 3:
                break

        if hasattr(node, "solutions"):
            for child in node.solutions.solution:
                self.ImportSolution(child, risk)
        return risk

    def ImportModule(self, node, survey):
        """
        Create a new :obj:`euphorie.content.module` object for a
        :obj:`euphorie.content.survey` given the details for a Module as an XML
        node.

        :returns: :obj:`euphorie.content.module`.
        """
        module = createContentInContainer(
            survey, "euphorie.module", title=str(node.title)
        )
        module.optional = node.get("optional") == "true"
        module.description = el_unicode(
            node, "description", is_etranslate_compatible=self.is_etranslate_compatible
        )
        module.external_id = attr_unicode(node, "external-id")
        if module.optional:
            module.question = str(node.question)
        module.solution_direction = el_unicode(node, "solution-direction")

        for child in node.iterchildren(tag=XMLNS + "risk"):
            self.ImportRisk(child, module)

        for child in node.iterchildren(tag=XMLNS + "module"):
            self.ImportModule(child, module)

        image = getattr(node, "image", None)
        if image is not None:
            (image, caption) = self.ImportImage(image)
            module.image = image
            module.caption = caption

        if node.get("hide_from_training"):
            behavior = IHideFromTraining(module, None)
            if behavior:
                behavior.hide_from_training = True

        return module

    def ImportProfileQuestion(self, node, survey):
        """
        Create a new :obj:`euphorie.content.profilequestion` object for a
        :obj:`euphorie.content.survey` given the details for a Profile Question
        as an XML node.

        :returns: :obj:`euphorie.content.profilequestion`.
        """
        type = node.get("type")
        if type == "optional":
            profile = createContentInContainer(
                survey, "euphorie.module", title=str(node.title)
            )
            profile.optional = True
        else:
            profile = createContentInContainer(
                survey, "euphorie.profilequestion", title=str(node.title)
            )
        profile.description = el_unicode(
            node, "description", is_etranslate_compatible=self.is_etranslate_compatible
        )
        profile.question = str(node.question)
        profile.external_id = attr_unicode(node, "external-id")
        for fname in ProfileQuestionLocationFields:
            setattr(profile, fname, el_unicode(node, fname.replace("_", "-")))
        profile.use_location_question = el_bool(node, "use-location-question", True)

        for child in node.iterchildren(tag=XMLNS + "risk"):
            self.ImportRisk(child, profile)

        for child in node.iterchildren(tag=XMLNS + "module"):
            self.ImportModule(child, profile)
        return profile

    def ImportTrainingQuestion(self, node, survey):
        """
        Create a new :obj:`euphorie.training_question` object for a
        :obj:`euphorie.training_question` given the details for a Training Question
        as an XML node.

        :returns: :obj:`euphorie.training_question`.
        """
        training_question = createContentInContainer(
            survey, "euphorie.training_question", title=str(node.title)
        )
        training_question.right_answer = str(node.right_answer)
        training_question.wrong_answer_1 = str(node.wrong_answer_1)
        training_question.wrong_answer_2 = str(node.wrong_answer_2)
        return training_question

    def ImportSurvey(self, node, group, version_title):
        """
        Create a new :obj:`euphorie.content.survey` object for a
        :obj:`euphorie.content.surveygroup` given the details for a Survey as an
        XML node.

        :returns: :obj:`euphorie.content.survey`.
        """
        survey = createContentInContainer(group, "euphorie.survey", title=version_title)
        survey.introduction = el_unicode(
            node, "introduction", is_etranslate_compatible=self.is_etranslate_compatible
        )
        survey.classification_code = el_unicode(node, "classification-code")
        survey.language = el_string(node, "language")
        tti = getUtility(IToolTypesInfo)
        survey.report_completion_threshold = attr_int(
            node, "report_completion_threshold", "value"
        )
        survey.enable_web_training = attr_bool(node, "enable_web_training", "value")
        survey.enable_test_questions = attr_bool(node, "enable_test_questions", "value")
        survey.enable_email_reminder = attr_bool(node, "enable_email_reminder", "value")
        survey.num_training_questions = attr_int(
            node, "num_training_questions", "value"
        )
        if self.is_etranslate_compatible:
            survey.tool_type = attr_string(
                node, "tool_type", "value", tti.default_tool_type
            )
            survey.measures_text_handling = attr_string(
                node, "measures_text_handling", "value", "full"
            )
            survey.integrated_action_plan = attr_bool(
                node, "integrated_action_plan", "value"
            )
            survey.evaluation_optional = attr_bool(node, "evaluation-optional", "value")
        else:
            survey.tool_type = el_string(node, "tool_type", tti.default_tool_type)
            survey.measures_text_handling = el_string(
                node, "measures_text_handling", "full"
            )
            survey.integrated_action_plan = el_bool(node, "integrated_action_plan")
            survey.evaluation_optional = el_bool(node, "evaluation-optional")
        survey.external_id = attr_unicode(node, "external-id")
        external_site_logo = getattr(node, "external_site_logo", None)
        if external_site_logo is not None:
            (image, caption) = self.ImportImage(external_site_logo)
            survey.external_site_logo = image
        lead_image = getattr(node, "image", None)
        if lead_image is not None:
            (image, caption) = self.ImportImage(lead_image)
            survey.image = image

        if IToolCategory.providedBy(survey):
            IToolCategory(survey).tool_category = [
                x.replace(COMMA_REPLACEMENT, ",").strip()
                for x in el_unicode(node, "tool-category", "").split(",")
            ]
        for child in node.iterchildren():
            if child.tag == XMLNS + "profile-question":
                self.ImportProfileQuestion(child, survey)
            elif child.tag == XMLNS + "module":
                self.ImportModule(child, survey)
            elif child.tag == XMLNS + "training_question":
                self.ImportTrainingQuestion(child, survey)
        return survey

    def __call__(
        self, input, surveygroup_title, survey_title, is_etranslate_compatible=False
    ):
        """Import a new survey from the XML data in `input` and create a new
        survey with the given `title`.

        `input` has to be either the raw XML input, or a
        `lxml.objectify` enablde DOM.
        """
        self.is_etranslate_compatible = is_etranslate_compatible
        if isinstance(input, (str,) + (bytes,)):
            try:
                sector = lxml.objectify.fromstring(safe_bytes(input))
            except Exception:
                # It might be that the file is huge, let's try a different approach
                parser = lxml.etree.XMLParser(huge_tree=True)
                sector_tree = lxml.objectify.parse(
                    BytesIO(safe_bytes(input)), parser=parser
                )
                sector = lxml.objectify.fromstring(lxml.etree.tostring(sector_tree))
        else:
            sector = input

        root = aq_inner(self.context)
        sg = createContentInContainer(
            root,
            "euphorie.surveygroup",
            title=surveygroup_title or str(sector.survey.title),
        )

        return self.ImportSurvey(sector.survey, sg, survey_title)


class SectorImporter(SurveyImporter):
    """:returns: :obj:`euphorie.content.survey` root"""

    def __call__(
        self, input, sector_title, sector_login, surveygroup_title, survey_title
    ):
        if isinstance(input, (str,) + (bytes,)):
            sector = lxml.objectify.fromstring(safe_bytes(input))
        else:
            sector = input

        country = aq_inner(self.context)

        if not sector_title:
            sector_title = str(sector.title)
        root = createContentInContainer(country, "euphorie.sector", title=sector_title)

        account = getattr(sector, "account", None)
        if account is not None:
            root.login = sector_login or account.get("login").lower()
            root.password = account.get("password")
        else:
            root.login = sector_login

        contact = getattr(sector, "contact", None)
        if contact is not None:
            root.contact_name = el_unicode(contact, "name")
            root.contact_email = el_string(contact, "email")

        logo = getattr(sector, "logo", None)
        if logo is not None:
            root.logo = self.ImportImage(logo)[0]

        if hasattr(sector, "survey"):
            sg = createContentInContainer(
                root,
                "euphorie.surveygroup",
                title=surveygroup_title or str(sector.survey.title),
            )
            self.ImportSurvey(sector.survey, sg, survey_title)

        return root


class ImportSurvey(AutoExtensibleForm, form.Form):
    """The upload view for a :obj:`euphorie.content.sector`

    View name: @@upload
    """

    schema = IImportSurvey
    ignoreContext = True
    form_name = _("Import OiRA Tool version")

    importer_factory = SurveyImporter

    @button.buttonAndHandler(_("Upload"))
    def handleUpload(self, action):
        (data, errors) = self.extractData()
        if errors:
            return
        input = data["file"].data
        importer = self.importer_factory(self.context)
        try:
            survey = importer(
                input,
                data["surveygroup_title"],
                data["survey_title"],
                data["is_etranslate_compatible"],
            )
        except lxml.etree.XMLSyntaxError:
            raise WidgetActionExecutionError(
                "file",
                Invalid(
                    _("error_invalid_xml", default="Please upload a valid XML file")
                ),
            )

        IStatusMessage(self.request).addStatusMessage(
            _("upload_success", default="Successfully imported the OiRA Tool"),
            type="success",
        )
        state = getMultiAdapter((survey, self.request), name="plone_context_state")
        self.request.response.redirect(state.view_url())

    @button.buttonAndHandler(_("button_cancel", default="Cancel"))
    def handleCancel(self, action):
        state = getMultiAdapter(
            (aq_inner(self.context), self.request), name="plone_context_state"
        )
        self.request.response.redirect(state.view_url())



from Acquisition import aq_parent
from euphorie.content.risk import EnsureInterface
from euphorie.content.risk import Risk
from euphorie.content.tests.utils import addSurvey
from euphorie.content.tests.utils import createSector
from euphorie.testing import EuphorieFunctionalTestCase
from euphorie.testing import EuphorieIntegrationTestCase
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD

import re


class RiskTests(EuphorieIntegrationTestCase):
    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createRisk(self, algorithm="kinney"):
        country = self.portal.sectors.nl
        sector = self._create(country, "euphorie.sector", "sector")
        surveygroup = self._create(
            sector, "euphorie.surveygroup", "group", evaluation_algorithm=algorithm
        )
        survey = self._create(surveygroup, "euphorie.survey", "survey")
        module = self._create(survey, "euphorie.module", "module")
        risk = self._create(module, "euphorie.risk", "risk")
        EnsureInterface(risk)
        return risk

    def testNotGloballyAllowed(self):
        self.loginAsPortalOwner()
        types = [fti.id for fti in self.portal.allowedContentTypes()]
        self.assertTrue("euphorie.risk" not in types)

    def testAllowedContentTypes(self):
        self.loginAsPortalOwner()
        risk = self.createRisk()
        types = [fti.id for fti in risk.allowedContentTypes()]
        self.assertEqual(set(types), {"euphorie.solution"})

    def testCanBeCopied(self):
        self.loginAsPortalOwner()
        risk = self.createRisk()
        self.assertTrue(risk.cb_isCopyable())

    def testDefaultEvaluationAlgorithm(self):
        risk = Risk()
        self.assertEqual(risk.evaluation_algorithm(), "kinney")

    def testFrenchEvaluationAlgorithm(self):
        self.loginAsPortalOwner()
        risk = self.createRisk("french")
        self.assertEqual(risk.evaluation_algorithm(), "french")


class RiskFunctionalTests(EuphorieFunctionalTestCase):
    _default_credentials = {
        "username": SITE_OWNER_NAME,
        "password": SITE_OWNER_PASSWORD,
    }

    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createRisk(self, algorithm="kinney"):
        country = self.portal.sectors.nl
        sector = self._create(country, "euphorie.sector", "sector")
        surveygroup = self._create(
            sector, "euphorie.surveygroup", "group", evaluation_algorithm=algorithm
        )
        survey = self._create(surveygroup, "euphorie.survey", "survey")
        module = self._create(survey, "euphorie.module", "module")
        risk = self._create(module, "euphorie.risk", "risk")
        EnsureInterface(risk)
        return risk

    def testDescriptionSanitised(self):
        self.loginAsPortalOwner()
        risk = self.createRisk()
        risk.title = "Risk title"
        risk.problem_description = "Problem description"
        browser = self.get_browser(logged_in=True)
        browser.open("%s/@@edit" % risk.absolute_url())
        browser.getControl(name="form.widgets.description").value = "Raw text"
        browser.handleErrors = False
        browser.getControl(name="form.buttons.save").click()
        self.assertEqual(risk.description, "<p>Raw text</p>")

    def testLegalReferenceSanitised(self):
        self.loginAsPortalOwner()
        risk = self.createRisk()
        risk.title = "Risk title"
        risk.description = "<p>Description</p>"
        risk.problem_description = "Problem description"
        browser = self.get_browser(logged_in=True)
        browser.open("%s/@@edit" % risk.absolute_url())
        browser.getControl(name="form.widgets.legal_reference").value = "Raw text"
        browser.getControl(name="form.buttons.save").click()
        self.assertEqual(risk.legal_reference, "<p>Raw text</p>")

    def testFrenchEvaluationOptionsShown(self):
        self.loginAsPortalOwner()
        risk = self.createRisk("french")
        risk.default_frequency = 9
        risk.default_severity = 7
        group = aq_parent(aq_parent(aq_parent(risk)))
        self.assertEqual(group.evaluation_algorithm, "french")
        browser = self.get_browser(logged_in=True)
        browser.handleErrors = False
        browser.open(risk.absolute_url())
        self.assertTrue("Severe" in browser.contents)
        self.assertTrue("Very often or regularly" in browser.contents)

    def testImageFromOtherSectorAccount(self):
        # http://code.simplon.biz/tracker/euphorie/ticket/143
        sector = createSector(self.portal, login="sector")
        self.loginAsPortalOwner()
        survey = addSurvey(sector)
        createSector(self.portal, id="sector2", login="sector2", password="sector2")
        self.logout()
        browser = self.get_browser()
        browser.open("%s/@@login" % self.portal.absolute_url())
        browser.getControl(name="__ac_name").value = "sector2"
        browser.getControl(name="__ac_password").value = "sector2"
        browser.getForm(id="loginForm").submit()
        risk = survey["1"]["2"]
        browser.open(risk.absolute_url())
        match = re.search(
            r'<div class="introduction">\s*<img[^>]+src="([^"]+)', browser.contents
        )
        self.assertTrue(match)
        image_url = match.group(1)
        browser.open(image_url)
        self.assertEqual(browser.isHtml, False)
        self.assertEqual(browser.headers.get_content_maintype(), "image")

    def testFixedPriorityForm(self):
        # See https://github.com/euphorie/Euphorie/pull/98
        self.loginAsPortalOwner()
        risk = self.createRisk()
        risk.title = "Risk title"
        risk.description = "<p>Description</p>"
        risk.problem_description = "Problem description"
        browser = self.get_browser(logged_in=True)
        browser.open("%s/@@edit" % risk.absolute_url())
        browser.getControl(name="form.widgets.evaluation_method").value = ["fixed"]
        browser.getControl(name="form.widgets.default_priority").value = ["low"]
        browser.getControl(name="form.widgets.fixed_priority").value = ["high"]
        browser.getControl(name="form.buttons.save").click()
        self.assertEqual(risk.fixed_priority, "high")


class ConstructionFilterTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()

    def _create(self, container, *args, **kwargs):
        newid = container.invokeFactory(*args, **kwargs)
        return getattr(container, newid)

    def createStructure(self):
        self.country = self.portal.sectors.nl
        self.sector = self._create(self.country, "euphorie.sector", "sector")
        self.surveygroup = self._create(self.sector, "euphorie.surveygroup", "group")
        self.survey = self._create(self.surveygroup, "euphorie.survey", "survey")
        self.module = self._create(self.survey, "euphorie.module", "module")

    def testPreventRiskIfModuleExists(self):
        self.createStructure()
        types = [fti.id for fti in self.module.allowedContentTypes()]
        self.assertTrue("euphorie.risk" in types)
        self._create(self.module, "euphorie.module", "module")
        types = [fti.id for fti in self.module.allowedContentTypes()]
        self.assertTrue("euphorie.risk" not in types)



from euphorie.client.model import Account
from euphorie.client.model import Session
from euphorie.client.model import SurveySession
from plone import api
from Products.CMFCore.ActionInformation import ActionInfo
from Products.CMFPlone.ActionsTool import ActionsTool
from Products.Five import BrowserView
from transaction import commit

import logging


logger = logging.getLogger(__name__)


class CleanUpLastModifierId(BrowserView):
    """Clean up last_modifier_id."""

    def __call__(self):
        logger.info("Cleaning up last_modifier_id")
        session = Session()
        obsolete_guest_users = (
            session.query(Account, SurveySession)
            .filter(Account.id == SurveySession.last_modifier_id)
            .filter(Account.id != SurveySession.account_id)
            .filter(Account.account_type == "guest")
        )

        with session.no_autoflush:
            for guest_user, assessment in obsolete_guest_users:
                assessment.last_modifier_id = assessment.account_id
                logger.info("Updated session %s", assessment.id)

                num_assessments = (
                    session.query(SurveySession)
                    .filter(SurveySession.account_id == guest_user.id)
                    .count()
                )
                if num_assessments == 0:
                    session.delete(guest_user)
                    logger.info("Deleted user %s", guest_user.id)
        commit()
        return "Done"


class AdminMaintenanceView(BrowserView):
    """View for maintenance tasks (@@admin-maintenance)."""

    @property
    def actions(self) -> tuple[ActionInfo]:
        """Return all the actions available in the `maintenance_actions` category."""
        portal_actions: ActionsTool = api.portal.get_tool("portal_actions")
        return portal_actions.listActionInfos(
            object=self.context, categories=["maintenance_actions"]
        )



from euphorie.client.authentication import addEuphorieAccountPlugin
from euphorie.client.authentication import EuphorieAccountPlugin
from plone import api

import logging


log = logging.getLogger(__name__)


def enable_plugin_and_move_to_top(pas, plugin):
    infos = [
        info
        for info in pas.plugins.listPluginTypeInfo()
        if plugin.testImplements(info["interface"])
    ]
    plugin.manage_activateInterfaces([info["id"] for info in infos])
    for info in infos:
        for i in range(len(pas.plugins.listPluginIds(info["interface"]))):
            pas.plugins.movePluginsUp(info["interface"], [plugin.getId()])


def add_account_plugin(pas):
    addEuphorieAccountPlugin(pas, "euphorie", "Euphorie account manager")
    enable_plugin_and_move_to_top(pas, pas.euphorie)


def setupVarious(context):
    site = api.portal.get()
    pas = site.acl_users
    if not pas.objectIds([EuphorieAccountPlugin.meta_type]):
        add_account_plugin(pas)
    ppr = api.portal.get_tool("portal_password_reset")
    ppr.setExpirationTimeout(0.5)



from euphorie.client import model
from euphorie.testing import EuphorieIntegrationTestCase
from z3c.saconfig import Session


class completion_percentage_tests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.session = Session()
        account = model.Account(loginname="jane", password="secret")
        self.session.add(account)
        self.survey = model.SurveySession(
            title="Survey", zodb_path="survey", account=account
        )
        self.session.add(self.survey)
        self.session.flush()

    def test_no_risks(self):
        self.assertEqual(self.survey.completion_percentage, 0)

    def test_half_complete(self):
        self.mod1 = self.survey.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        self.q1 = self.mod1.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        self.q2 = self.mod1.addChild(
            model.Risk(title="question 2", risk_id="2", zodb_path="a/c")
        )
        self.session.flush()

        self.q1.postponed = False
        self.q1.identification = "no"

        self.assertEqual(self.survey.completion_percentage, 50)

    def test_two_thirds_complete(self):
        self.mod1 = self.survey.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        self.q1 = self.mod1.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        self.q2 = self.mod1.addChild(
            model.Risk(title="question 2", risk_id="2", zodb_path="a/c")
        )
        self.mod2 = self.survey.addChild(
            model.Module(title="module 2", module_id="2", zodb_path="k")
        )
        self.q21 = self.mod1.addChild(
            model.Risk(title="question 3", risk_id="3", zodb_path="k/l")
        )
        self.session.flush()

        self.q1.postponed = False
        self.q1.identification = "n/a"
        self.q21.postponed = False
        self.q21.identification = "yes"

        self.assertEqual(self.survey.completion_percentage, 67)

    def test_optional_module(self):
        self.mod1 = self.survey.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        self.q1 = self.mod1.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        self.mod2 = self.survey.addChild(
            model.Module(title="module 2", module_id="2", zodb_path="k")
        )
        self.q2 = self.mod2.addChild(
            model.Risk(title="question 2", risk_id="2", zodb_path="k/c")
        )
        self.q1.postponed = False
        self.q1.identification = "yes"
        self.mod2.skip_children = True

        self.assertEqual(self.survey.completion_percentage, 100)

    def test_optional_submodule(self):
        self.mod1 = self.survey.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        self.q1 = self.mod1.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        self.mod2 = self.survey.addChild(
            model.Module(title="module 2", module_id="2", zodb_path="k")
        )
        self.mod21 = self.mod2.addChild(
            model.Module(title="module 3", module_id="3", zodb_path="k/c")
        )
        self.q2 = self.mod21.addChild(
            model.Risk(title="question 2", risk_id="2", zodb_path="k/c/u")
        )
        self.q1.postponed = False
        self.q1.identification = "yes"
        self.mod21.skip_children = True

        self.assertEqual(self.survey.completion_percentage, 100)

    def test_optional_module_with_submodule(self):
        self.mod1 = self.survey.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        self.q1 = self.mod1.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        self.mod2 = self.survey.addChild(
            model.Module(title="module 2", module_id="2", zodb_path="k")
        )
        self.mod21 = self.mod2.addChild(
            model.Module(title="module 3", module_id="3", zodb_path="k/c")
        )
        self.q2 = self.mod21.addChild(
            model.Risk(title="question 2", risk_id="2", zodb_path="k/c/u")
        )

        self.q1.postponed = False
        self.q1.identification = "yes"
        self.mod2.skip_children = True
        self.mod21.skip_children = False
        self.assertEqual(self.survey.completion_percentage, 100)



from AccessControl import getSecurityManager
from Acquisition import aq_base
from Acquisition import aq_chain
from Acquisition import aq_inner
from Acquisition import aq_parent
from datetime import datetime
from dateutil import tz
from decimal import Decimal
from euphorie import MessageFactory as _
from euphorie.client import config
from euphorie.client.adapters.session_traversal import ITraversedSurveySession
from euphorie.client.client import IClient
from euphorie.client.country import IClientCountry
from euphorie.client.model import get_current_account
from euphorie.client.model import Group
from euphorie.client.model import OrganisationMembership
from euphorie.client.model import Session
from euphorie.client.model import SurveySession
from euphorie.client.sector import IClientSector
from euphorie.client.update import wasSurveyUpdated
from euphorie.client.utils import getSecret
from euphorie.content.survey import ISurvey
from euphorie.content.utils import getRegionTitle
from euphorie.content.utils import StripMarkup
from functools import cached_property
from json import dumps
from logging import getLogger
from os import path
from pkg_resources import resource_listdir
from plone import api
from plone.i18n.interfaces import ILanguageUtility
from plone.i18n.normalizer import idnormalizer
from plone.memoize import forever
from plone.memoize.instance import memoize
from plone.memoize.view import memoize_contextless
from plone.registry.interfaces import IRegistry
from plonetheme.nuplone.utils import isAnonymous
from Products.CMFCore.interfaces import ISiteRoot
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import ISecuritySchema
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from sqlalchemy import and_
from urllib.parse import urlencode
from user_agents import parse
from ZODB.POSException import POSKeyError
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.component.hooks import getSite
from zope.deprecation import deprecate
from zope.i18n import translate
from zope.i18nmessageid import MessageFactory


pl_message = MessageFactory("plonelocales")
pae_message = MessageFactory("plone.app.event")
logger = getLogger(__name__)


# XXX should to to config (registry?)
NAME_TO_PHASE = {
    "start": "preparation",
    "involve": "involve",
    "profile": "preparation",
    "identification": "identification",
    "customization": "identification",
    "actionplan": "actionplan",
    "report": "report",
    "consultancy": "consultancy",
    "consultants": "consultancy",
    "status": "status",
    "help": "help",
    "new-email": "useraction",
    "account-settings": "useraction",
    "account-delete": "useraction",
    "update": "preparation",
    "disclaimer": "help",
    "terms-and-conditions": "help",
    "training": "training",
    "email-reminder": "reminder",
}


class WebHelpers(BrowserView):
    """Browser view with utility methods that can be used in templates.

    Several methods in this view assume that the current survey can be
    found as an attribute on the request. This is normally setup by the
    :py:class:`euphorie.client.survey.SurveyPublishTraverser` traverser.

    View name: @@webhelpers
    """

    certificates_path = "++resource++euphorie.resources/assets/oira/certificates"
    media_path = "++resource++euphorie.resources/media"
    style_path = "++resource++euphorie.resources/assets/oira/style"
    script_path = "++resource++patternslib"

    brand = "oira"

    css_path = "++resource++euphorie.resources/assets/{brand}/style/all.css"
    css_path_min = "++resource++euphorie.resources/assets/{brand}/style/all.css"

    js_name = "bundle.min.js"

    favicon_path = (
        "++resource++euphorie.resources/assets/{brand}/favicon/apple-touch-icon.png"
    )

    group_model = Group
    hide_organisation_tab = False
    survey_session_model = SurveySession
    dashboard_tabs = ["surveys", "assessments", "certificates", "organisation"]

    navigation_tree_legend = [
        {"class": "unvisited", "title": _("Unvisited")},
        {"class": "postponed", "title": _("Postponed")},
        {"class": "answered", "title": _("Risk not present")},
        {"class": "answered risk", "title": _("Risk present")},
    ]

    # Inspection is meant for showing an edit-form in read-only/display mode.
    allow_inspecting_archived_sessions = False
    allow_inspecting_locked_sessions = False

    def to_decimal(self, value):
        """Transform value in to a decimal."""
        return Decimal(value)

    @property
    @memoize
    def resources_timestamp(self):
        return api.portal.get_registry_record(
            "euphorie.deployment.resources_timestamp", default=""
        )

    @property
    @memoize
    def server_timezone(self):
        return datetime.now(tz.tzlocal()).tzname()

    @property
    @memoize
    def sector(self):
        for obj in aq_chain(aq_inner(self.context)):
            if IClientSector.providedBy(obj):
                return obj

    @property
    @memoize
    def debug_mode(self):
        return api.env.debug_mode()

    @property
    @memoize
    def allow_social_sharing(self):
        return api.portal.get_registry_record(
            "euphorie.allow_social_sharing", default=False
        )

    @property
    @memoize
    def allow_guest_accounts(self):
        return api.portal.get_registry_record(
            "euphorie.allow_guest_accounts", default=False
        )

    @property
    @memoize
    def allow_self_registration(self):
        registry = getUtility(IRegistry)
        security_settings = registry.forInterface(ISecuritySchema, prefix="plone")
        return security_settings.enable_self_reg

    @property
    @memoize
    def use_involve_phase(self):
        return api.portal.get_registry_record(
            "euphorie.use_involve_phase", default=False
        )

    @property
    @memoize
    def use_consultancy_phase(self):
        return self.content_country_obj.enable_consultancy

    @property
    @memoize
    def use_training_module(self):
        globally_enabled = api.portal.get_registry_record(
            "euphorie.use_training_module", default=False
        )
        if not globally_enabled:
            return False
        country_enabled = self.content_country_obj.enable_web_training
        if not country_enabled:
            return False
        if self._survey is None:
            return True
        return self._survey.enable_web_training

    @property
    def display_training_module(self):
        if not self.use_training_module:
            return False
        traversed_session = self.traversed_session
        if traversed_session is None:
            return False

        training_slide = api.content.get_view(
            name="training_slide", context=self.context, request=self.request
        )
        if (
            training_slide.existing_measures_training
            or training_slide.planned_measures_training
        ):
            return True
        return False

    @property
    @memoize
    @deprecate(
        "Publication has been changed to locking. Deprecated in version 15.0.0.dev0"
    )
    def use_publication_feature(self):
        return api.portal.get_registry_record(
            "euphorie.use_locking_feature", default=False
        )

    @memoize
    @deprecate(
        "Publication has been changed to locking. Deprecated in version 15.0.0.dev0"
    )
    def use_publication_feature_for_session(self, session):
        # to be overwritten as needed
        return self.use_locking_feature

    @property
    @memoize
    def use_locking_feature(self):
        return api.portal.get_registry_record(
            "euphorie.use_locking_feature", default=False
        )

    @memoize
    def use_locking_feature_for_session(self, session):
        # to be overwritten as needed
        return self.use_locking_feature

    @property
    @memoize
    def use_archive_feature(self):
        return api.portal.get_registry_record(
            "euphorie.use_archive_feature", default=False
        )

    @property
    @memoize
    def use_help_section(self):
        return api.portal.get_registry_record("euphorie.use_help_section", default=True)

    # Feature switch, can be overwritten in subclass
    show_completion_percentage = False

    @property
    @memoize
    def show_certificates_tab(self):
        return self.use_training_module

    @property
    @memoize
    def default_country(self):
        return api.portal.get_registry_record("euphorie.default_country", default="")

    @property
    @memoize
    @deprecate("Use current_account instead. Deprecated in version 15.0.0.dev0")
    def _user(self):
        return self.current_account

    @property
    @memoize
    def current_account(self):
        return self.get_current_account()

    @property
    @memoize
    def anonymous(self):
        return isAnonymous(self.current_account)

    @property
    @memoize
    def is_guest_account(self):
        account = getattr(self.current_account, "account_type", None)
        return account == config.GUEST_ACCOUNT

    @property
    @memoize
    def session(self):
        raise Exception("Use the traversed session")

    @property
    @memoize
    def traversed_session(self):
        for obj in self.context.aq_chain:
            if ITraversedSurveySession.providedBy(obj):
                return obj

    def redirectOnSurveyUpdate(self):
        """Utility method for views to check if a survey has been updated, and
        if so redirect the user to the update confirmation page is generated.

        The return value is `True` if an update is required and `False`
        otherwise.
        """
        if not self.can_edit_session:
            return False
        traversed_session = self.traversed_session
        session = traversed_session.session
        survey = traversed_session.aq_parent
        if not wasSurveyUpdated(session, survey):
            return False
        self.request.response.redirect(
            "{session_url}/@@update?initial_view=1".format(
                session_url=traversed_session.absolute_url()
            )
        )
        return True

    @property
    @memoize
    def session_id(self):
        traversed_session = self.traversed_session
        return traversed_session.session.id if traversed_session else ""

    @memoize
    @deprecate(
        "Replace with session.completion_percentage "
        "or self.traversed_session.session.completion_percentage. "
        "Deprecated in version 14.1.4.dev0"
    )
    def update_completion_percentage(self, session=None):
        if not session:
            session = self.traversed_session.session
        return session.completion_percentage

    def get_progress_indicator_title(self, completion_percentage=None):
        if completion_percentage is None and self.traversed_session is not None:
            completion_percentage = self.traversed_session.session.completion_percentage
        title = _(
            "progress_indicator_title",
            default="${completion_percentage}% Complete",
            mapping={"completion_percentage": completion_percentage or 0},
        )
        return api.portal.translate(title)

    @memoize
    def get_came_from(self, default=None):
        came_from = self.request.form.get("came_from")
        if not came_from:
            return default
        if isinstance(came_from, list):
            # If came_from is both in the querystring and the form data
            came_from = came_from[0]
        put = getToolByName(self.context, "portal_url")
        if came_from and not put.isURLInPortal(came_from):
            return default
        return came_from

    @property
    @memoize
    def came_from(self):
        default = aq_parent(self.context).absolute_url()
        came_from = self.get_came_from(default=default)
        return came_from

    @property
    @memoize
    def country_obj(self):
        for obj in aq_chain(aq_inner(self.context)):
            if IClientCountry.providedBy(obj):
                return obj

    @property
    @memoize
    def county_obj_via_parents(self):
        for obj in self.request.PARENTS:
            if IClientCountry.providedBy(obj):
                return obj

    @property
    @memoize
    def country_name(self):
        obj = self.country_obj
        if not obj:
            return ""
        return obj.Title()

    @property
    @memoize
    def content_country_obj(self):
        """Return the country object from the content area."""
        country_id = self.country_obj.id
        root = getUtility(ISiteRoot)
        country = getattr(root.sectors, country_id)
        return country

    @property
    @memoize
    def selected_country(self):
        """Return the country id that is present in the path.

        Fall back to the default_country
        """
        country = self.county_obj_via_parents
        if country:
            return country.getId()
        return self.default_country

    @property
    @memoize
    def sector_name(self):
        for obj in aq_chain(aq_inner(self.context)):
            if IClientSector.providedBy(obj):
                return obj.Title()
        return ""

    @property
    @memoize
    def tool_name(self):
        obj = self._survey
        if not obj:
            return ""
        return obj.Title()

    @property
    @memoize
    def tool_description(self):
        obj = self._survey
        if not obj:
            return ""
        ploneview = self.context.restrictedTraverse("@@plone")
        return ploneview.cropText(StripMarkup(obj.introduction), 800)

    @property
    @memoize
    def language_code(self):
        lt = getToolByName(self.context, "portal_languages")
        lang = lt.getPreferredLanguage()
        return lang

    @memoize_contextless
    def getNameForLanguageCode(self, langCode, native=True):
        lang_util = getUtility(ILanguageUtility)
        info = lang_util.getAvailableLanguageInformation().get(langCode, None)
        if info is not None:
            if native:
                return info.get("native", info.get("name", None))
            lang_names = self.request.locale.displayNames.languages
            name = lang_names.get(langCode, langCode)
            # Better show the (English) full name than only the language code
            if name == langCode:
                name = info.get("name")
            return name.capitalize()
        return None

    @memoize
    def getTranslatedCountryName(self, country_id):
        return getRegionTitle(self.request, country_id)

    @property
    @forever.memoize
    def available_help_languages(self):
        exclude = {"illustrations"}
        return (
            set(resource_listdir("plonestatic.euphorie", "resources/assets/oira/help"))
            - exclude
        )

    @property
    @memoize
    def help_language(self):
        lang = self.language_code
        # No country-specific Help texts are curently supported
        if lang.find("-"):
            lang = lang.split("-")[0]
        return lang if lang in self.available_help_languages else "en"

    def get_username(self):
        member = api.user.get_current()
        return member.getProperty("fullname") or member.getUserName()

    def get_webstats_js(self):
        return api.portal.get_registry_record("plone.webstats_js")

    def language_dict(self):
        site = getSite()
        ltool = getToolByName(site, "portal_languages")
        return ltool.getAvailableLanguages()

    @property
    def macros(self):
        return self.index.macros

    @property
    @memoize
    def country(self):
        """XXX it would be better to write this country id, consider
        deprecating this."""
        obj = self.country_obj
        if not obj:
            return ""
        return obj.id

    def logoMode(self):
        return "alien" if "alien" in self.extra_css else "native"

    def check_markup(self, text):
        if StripMarkup(text).strip():
            return text

    @property
    @memoize
    def extra_css(self):
        sector = self.sector
        if sector is None:
            return ""

        sector = aq_base(sector)
        parts = []

        if getattr(sector, "logo", None) is not None:
            parts.append("alien")

        lt = getToolByName(self.context, "portal_languages")
        lang = lt.getPreferredLanguage()
        parts.append("language-%s" % lang)

        return " " + " ".join(parts)

    @property
    @memoize
    def sector_title(self):
        """Return the title to use for the current sector.

        If the current context is not in a sector return the agency name
        instead.
        """
        sector = self.sector
        if sector is not None and getattr(aq_base(sector), "logo", None) is not None:
            return sector.Title()
        else:
            return _(
                "title_tool",
                default="OiRA - Online interactive Risk Assessment",
            )

    @property
    @memoize
    def client(self):
        if not self.context:
            return
        for obj in self.context.aq_chain:
            if IClient.providedBy(obj):
                return obj

    @property
    @memoize
    def client_url(self):
        """Return the absolute URL for the client."""
        return self.client.absolute_url() if self.client else "."

    @property
    @memoize
    def portal_url(self):
        return api.portal.get().absolute_url()

    @property
    @memoize
    def country_or_client_url(self):
        """Return the country URL, but fall back to the client URL in case the
        country URL is None.

        Relevant in tests only, as far as I can see
        """
        return self.country_url or self.client_url

    def _base_url(self):
        """Return a base URL to be used for non-survey specific pages.

        If we are in a survey the help page will be located there.
        Otherwise the country will be used as parent.
        """
        base_url = self.survey_url()
        if base_url is not None and aq_inner(self.context).absolute_url().startswith(
            base_url
        ):
            return base_url
        base_url = self.country_url
        if base_url is not None:
            return base_url
        return self.client_url

    @property
    @memoize
    def base_url(self):
        if self.anonymous:
            base_url = self.country_url
            if base_url is not None:
                return base_url
            return self.client_url
        return self._base_url()

    @property
    @memoize
    def certificates_url(self):
        return f"{self.client_url}/{self.certificates_path}"

    @property
    @memoize
    def media_url(self):
        return f"{self.client_url}/{self.media_path}"

    @property
    @memoize
    def style_url(self):
        return f"{self.client_url}/{self.style_path}"

    @property
    @memoize
    def css_url(self):
        return "{}/{}?t={}".format(
            self.client_url,
            (
                self.css_path.format(brand=self.brand)
                if not self.debug_mode
                else self.css_path_min.format(brand=self.brand)
            ),
            self.resources_timestamp,
        )

    @property
    @memoize
    def js_url(self):
        return "{}/{}/{}?t={}".format(
            self.client_url,
            self.script_path,
            self.js_name,
            self.resources_timestamp,
        )

    @property
    @memoize
    def favicon_url(self):
        return "{}/{}?t={}".format(
            self.client_url,
            self.favicon_path.format(brand=self.brand),
            self.resources_timestamp,
        )

    @property
    @memoize
    @deprecate(
        "Replace with  a check for `webhelpers.survey_url`, "
        "deprecated in version 11.1.2"
    )
    def is_outside_of_survey(self):
        if self._base_url() != self.survey_url():
            return True
        if (
            self.request.get("ACTUAL_URL").split("/")[-1]
            == self.survey_url().split("/")[-1]
        ):
            return True
        return False

    @property
    @memoize
    def get_survey_title(self):
        survey = self._survey
        if not survey:
            return None
        if getattr(self, "session", None) and "/".join(
            survey.getPhysicalPath()
        ).endswith(self.session.zodb_path):
            return self.session.title
        return survey.title

    def get_phase(self):
        head, tail = path.split(self.request.PATH_INFO)
        while tail:
            tail = tail.replace("@", "").split("_")[0]
            if tail in NAME_TO_PHASE:
                return NAME_TO_PHASE[tail]
            head, tail = path.split(head)
        return ""

    @cached_property
    def phase(self):
        return self.get_phase()

    def get_dashboard_tab(self):
        tail = path.split(self.request.PATH_INFO.rstrip("/"))[-1].lstrip("@@")
        if tail in self.dashboard_tabs:
            return tail
        return "dashboard"

    @property
    def came_from_param(self):
        if self.came_from:
            # If the tool has a notification message, we cannot allow to deeplink
            # into it, since the user might then miss the notificaton.
            if not self.tool_notification():
                return urlencode({"came_from": self.came_from})
            survey_url = self.survey_url()
            if survey_url:
                return urlencode({"came_from": survey_url})
        return ""

    @property
    @memoize
    def help_url(self):
        """Return the URL to the current online help page.

        If we are in a survey the help page will be located there.
        Otherwise the country will be used as parent.
        """
        return "%s/help" % self._base_url()

    @property
    @memoize
    def about_url(self):
        """Return the URL to the current online about page.

        If we are in a survey the help page will be located there.
        Otherwise the country will be used as parent.
        """
        return "%s/about" % self._base_url()

    @property
    @memoize
    def authenticated(self):
        """Check if the current user is authenticated."""
        user = getSecurityManager().getUser()
        return user is not None and user.getUserName() != "Anonymous User"

    @property
    @memoize
    def country_url(self):
        """Return the absolute URL for country page."""
        sector = self.sector
        if sector is not None:
            return aq_parent(sector).absolute_url()

        for parent in aq_chain(aq_inner(self.context)):
            if IClientCountry.providedBy(parent):
                return parent.absolute_url()

        return None

    @property
    @memoize
    def session_overview_url(self):
        """Return the absolute URL for the session overview."""
        return self.country_url

    @property
    @memoize
    def sector_url(self):
        """Return the URL for the current survey."""
        sector = self.sector
        if sector is None:
            return None
        return sector.absolute_url()

    @property
    @memoize
    def _survey(self):
        for parent in aq_chain(aq_inner(self.context)):
            if ISurvey.providedBy(parent):
                return parent

    @property
    @memoize
    def is_survey(self):
        """Return `True` if the webhelper's context is within a survey, `False`
        otherwise.
        """
        return bool(self._survey)

    @memoize
    def survey_url(self, phase=None):
        """Return the URL for the curreny survey.

        If a phase is specified the URL for the first page of that phase
        is returned.
        """
        survey = self._survey
        if not survey:
            return None

        url = survey.absolute_url()
        if phase is not None:
            url += "/@@%s" % phase
        return url

    @memoize
    def survey_zodb_path(self):
        """Construct zodb_path from current survey.

        Helper method, since I don't always trust self.session.zodb_path
        """
        elems = []
        obj = self._survey
        if not obj:
            return ""
        while not IClient.providedBy(obj):
            elems.append(obj.id)
            obj = aq_parent(obj)
        elems.reverse()
        return "/".join(elems)

    @property
    @memoize
    def integrated_action_plan(self):
        if not api.portal.get_registry_record(
            "euphorie.use_integrated_action_plan", default=False
        ):
            return False
        return getattr(self._survey, "integrated_action_plan", False)

    @property
    @memoize
    def use_action_plan_phase(self):
        return not self.integrated_action_plan

    @property
    @memoize
    def report_completion_threshold(self):
        survey = self._survey
        if not survey:
            return None
        return survey.report_completion_threshold or 0

    @property
    @memoize
    def use_email_reminder(self):
        survey = self._survey
        if not survey:
            return None
        return survey.enable_email_reminder

    @property
    @memoize
    def in_session(self):
        """Check if there is an active survey session."""
        return self._survey is not None

    def is_initialised_session(self, session):
        """A session without children has not passed the Identification phase
        yet."""
        return session.children().count()

    @property
    @memoize
    def appendix_documents(self):
        """Return a list of items to be shown in the appendix."""
        documents = api.portal.get().documents

        lt = getToolByName(self.context, "portal_languages")
        lang = lt.getPreferredLanguage()
        if "-" in lang:
            languages = [lang, lang.split("-")[0]]
        else:
            languages = [lang]

        for lang in languages:
            docs = documents.get(lang, None)
            if docs is None:
                continue
            appendix = docs.get("appendix", None)
            if appendix is not None:
                break
        else:
            return []

        base_url = self._base_url()
        return [
            {"url": f"{base_url}/appendix/{page.id}", "title": page.Title()}
            for page in appendix.values()
        ]

    @property
    @memoize
    def is_iphone(self):
        """Check if the current request is from an iPhone or similar device
        (such as an iPod touch)."""
        agent = self.request.get_header("User-Agent", "")
        return "iPhone" in agent

    def months(self, length="wide"):
        calendar = self.request.locale.dates.calendars["gregorian"]
        months = calendar.monthContexts["format"].months[length]
        return sorted(months.items())

    def timezoned_date(self, mydate=None):
        if mydate is None:
            return None
        # If the date already has a timezone, don't touch it
        if mydate.tzinfo is not None:
            return mydate
        local_tz = tz.gettz(self.server_timezone)
        return mydate.replace(tzinfo=local_tz)

    @property
    @memoize
    def get_sector_logo(self):
        sector = self.sector
        if sector is None:
            return None
        real_sector = sector._sector()
        images = getMultiAdapter((real_sector, self.request), name="images")
        try:
            return images.scale("logo", height=400, direction="up") or None
        except POSKeyError:
            return None

    @memoize
    def get_tool_image_url(self, survey=None):
        if not survey:
            survey = self._survey
        if getattr(survey, "image", None):
            return f"{survey.absolute_url()}/@@images/image/large"

    def messages(self):
        status = IStatusMessage(self.request)
        messages = status.show()
        for m in messages:
            m.id = idnormalizer.normalize(m.message)
        return messages

    def _getLanguages(self):
        lt = getToolByName(self.context, "portal_languages")
        lang = lt.getPreferredLanguage()
        if "-" in lang:
            return [lang, lang.split("-")[0], "en"]
        else:
            return [lang, "en"]

    def _findMOTD(self):
        documents = api.portal.get().documents

        motd = None
        for lang in self._getLanguages():
            docs = documents.get(lang, None)
            if docs is None:
                continue
            motd = docs.get("motd", None)
            if motd is not None:
                return motd

    def splash_message(self):
        message = None
        motd = self._findMOTD()
        if motd:
            now = datetime.now()
            message = dict(
                title=StripMarkup(motd.description),
                text=motd.body,
                id="motd{}{}".format(
                    motd.modification_date.strftime("%Y%m%d%H%M%S"),
                    now.strftime("%Y%m%d"),
                ),
            )
        return message

    def tool_notification(self):
        message = None
        obj = self._survey
        if ISurvey.providedBy(obj) and obj.hasNotification():
            now = datetime.now()
            message = dict(
                title=obj.tool_notification_title,
                text=obj.tool_notification_message,
                id="tool{}{}{}".format(
                    obj.modification_date.strftime("%Y%m%d%H%M%S"),
                    now.strftime("%Y%m%d"),
                    "".join(obj.getPhysicalPath()[2:]),
                ),
            )
        return message

    def closetext(self):
        return api.portal.translate(_("button_close", default="Close"))

    def email_sharing_text(self):
        return api.portal.translate(_("I wish to share the following with you"))

    def getSecret(self):
        return getSecret()

    @memoize_contextless
    def get_current_account(self):
        return get_current_account()

    @memoize_contextless
    def get_my_group_id(self):
        account = self.get_current_account()
        return account and account.group_id or ""

    @memoize
    def get_session_group_id(self):
        return getattr(self.session, "group_id", "")

    @property
    @memoize
    def can_view_session(self):
        account = self.get_current_account()
        if not account:
            return False
        session = self.traversed_session.session

        # Check if it is my session
        if session.account_id == account.id:
            return True

        # Check if it is a session of my organisation
        if (
            Session()
            .query(OrganisationMembership)
            .filter(
                and_(
                    OrganisationMembership.owner_id == session.account_id,
                    OrganisationMembership.member_id == account.id,
                )
            )
            .count()
        ):
            return True
        return session in account.acquired_sessions

    @property
    @memoize
    def can_edit_session(self):
        if not self.can_view_session:
            return False
        session = self.traversed_session.session
        if session.is_archived:
            return False
        if session.is_locked:
            return False
        return True

    @cached_property
    def can_inspect_session(self):
        """Can the user inspect the session?

        If a session is archived or locked, you cannot edit it.
        This used to mean that you cannot see the answers to risks, or indeed
        the entire 'identification' phase.  We may want to allow inspecting
        those parts without being able to edit them.  In the UI we could show
        the chosen answers for risks, without showing an editable form.
        """
        if not self.can_view_session:
            return False
        session = self.traversed_session.session
        if session.is_archived and not self.allow_inspecting_archived_sessions:
            return False
        if session.is_locked and not self.allow_inspecting_locked_sessions:
            return False
        return True

    @property
    @deprecate(
        "Publication has been changed to locking. Deprecated in version 15.0.0.dev0"
    )
    def can_publish_session(self):
        return self.can_lock_session

    @property
    @memoize
    def can_archive_session(self):
        if not self.use_archive_feature:
            return False
        if self.traversed_session.session.is_archived:
            return False
        return self.can_edit_session

    @property
    @memoize
    def can_delete_session(self):
        return self.can_edit_session

    @property
    @memoize
    def is_owner(self):
        """Check if the current user is the owner of the session."""
        session = self.traversed_session.session
        return self.get_current_account() == session.account

    @property
    @memoize
    def is_manager(self):
        """Check if the current user is a manager of the session."""
        account = self.get_current_account()
        if not account:
            return False
        session = self.traversed_session.session
        if (
            Session()
            .query(OrganisationMembership)
            .filter(
                and_(
                    OrganisationMembership.owner_id == session.account_id,
                    OrganisationMembership.member_id == account.id,
                    OrganisationMembership.member_role == "manager",
                )
            )
            .count()
        ):
            return True

    @property
    @memoize
    def can_lock_session(self):
        if not self.use_locking_feature:
            return False
        if self.is_owner:
            return True
        if self.is_manager:
            return True
        return False

    @property
    @memoize
    def can_unlock_session(self):
        return self.can_lock_session

    def resume(self, session):
        """Resume a session for the current user if he is allowed to."""
        raise Exception("Obsolete, we traverse to sessions now")

    def as_md(self, text):
        """Return a text with Carriage Returns formatted as a Markdown."""
        return "\r\n".join([x for x in text.split("\r")])

    def show_logo(self):
        """In plain Euphorie, the logo is always shown."""
        return True

    @property
    @memoize
    def portal_transforms(self):
        return api.portal.get_tool("portal_transforms")

    def get_safe_html(self, text):
        if not text:
            return ""
        data = self.portal_transforms.convertTo(
            "text/x-html-safe", text, mimetype="text/html"
        )
        return data.getData()

    @property
    @memoize
    def custom_js(self):
        """Return custom JavaScript where necessary."""
        return ""

    @memoize_contextless
    def date_picker_i18n_json(self):
        """Taken from: https://github.com/ploneintranet/ploneintranet/blob/mast
        er/src/ploneintranet/layout/browser/date_picker.py  # noqa: E501.

        Use this like: <input class="pat-date-picker"        ... data-
        pat-date-picker="...; i18n: ${portal_url}/@@date-
        picker-i18n.json; ..."        />
        """
        json = dumps(
            {
                "previousMonth": api.portal.translate(pae_message("prev_month_link")),
                "nextMonth": api.portal.translate(pae_message("next_month_link")),
                "months": [
                    api.portal.translate(pl_message(month))
                    for month in [
                        "month_jan",
                        "month_feb",
                        "month_mar",
                        "month_apr",
                        "month_may",
                        "month_jun",
                        "month_jul",
                        "month_aug",
                        "month_sep",
                        "month_oct",
                        "month_nov",
                        "month_dec",
                    ]
                ],
                "weekdays": [
                    api.portal.translate(pl_message(weekday))
                    for weekday in [
                        "weekday_sun",
                        "weekday_mon",
                        "weekday_tue",
                        "weekday_wed",
                        "weekday_thu",
                        "weekday_fri",
                        "weekday_sat",
                    ]
                ],
                "weekdaysShort": [
                    api.portal.translate(pl_message(weekday_abbr))
                    for weekday_abbr in [
                        "weekday_sun_abbr",
                        "weekday_mon_abbr",
                        "weekday_tue_abbr",
                        "weekday_wed_abbr",
                        "weekday_thu_abbr",
                        "weekday_fri_abbr",
                        "weekday_sat_abbr",
                    ]
                ],
            }
        )
        return json

    def get_sessions_query(
        self,
        context=None,
        searchable_text=None,
        order_by=False,
        include_archived=False,
        filter_by_account=True,
        filter_by_group=False,
        table=None,
    ):
        """Method to return a query that looks for sessions.

        :param context: limits the sessions under this context
        :param searchable_text: filters on survey title
        :param order_by: is by default on survey title but you can pass
            None to disable ordering or a SQLAlchemy expression
        :param include_archived: unless explicitely set to a truish value,
            archived sessions are excluded
        :param filter_by_account: True means current account.
            A falsish value means do not filter.
            Otherwise try to interpret the user input:
            a string or an int means the account_id should be that value,
            an object account will be used to extract the account id,
            from an iterable we will try to extract the account ids
        :param filter_by_group: True means current account group,
            A falsish value means do not filter.
            Otherwise try to interpret the user input:
            a string or an int means the group_id should be that value,
            an object group will be used to extract the group id,
            and from an iterable we will try to extract the group ids
        :param table: if None use the default table in `cls.survey_session_model`
            otherwise use the one passed by the user
        """
        if table is None:
            table = self.survey_session_model
        query = Session.query(table)

        query = query.filter(table.get_context_filter(context or self.context))

        if filter_by_account:
            query = query.filter(table.get_account_filter(account=filter_by_account))
        if filter_by_group:
            query = query.filter(table.get_group_filter(group=filter_by_group))
        if not include_archived:
            query = query.filter(table.get_archived_filter())
        if searchable_text:
            query = query.filter(table.title.ilike(searchable_text))
        if order_by is None:
            pass  # Do not append any order_by
        elif order_by:
            query = query.order_by(order_by)
        else:
            query = query.order_by(table.modified.desc(), table.title)
        return query

    @memoize
    def is_outdated_browser(self):
        ua_string = self.request.get("HTTP_USER_AGENT", "")
        ua = parse(ua_string)
        return ua.browser.family == "IE"

    @property
    def pat_validation_messages(self):
        lang = getattr(self.request, "LANGUAGE", "en")
        if "-" in lang:
            elems = lang.split("-")
            lang = f"{elems[0]}_{elems[1].upper()}"
        messages = {
            "message-date": translate(
                _("error_validation_date", default="This value must be a valid date"),
                target_language=lang,
            ),
            "message-datetime": translate(
                _(
                    "error_validation_datetime",
                    default="This value must be a valid date and time",
                ),
                target_language=lang,
            ),
            "message-email": translate(
                _(
                    "error_validation_email",
                    default="This value must be a valid email address",
                ),
                target_language=lang,
            ),
            "message-number": translate(
                _("error_validation_number", default="This value must be a number"),
                target_language=lang,
            ),
            "message-required": translate(
                _("message_field_required", default="Please fill out this field."),
                target_language=lang,
            ),
        }
        return "; ".join([f"{key}: {value}" for key, value in messages.items()])

    @property
    @memoize
    def is_new_session(self):
        if self.request.get("new_session"):
            return True
        session = self.traversed_session.session
        return session.children().count() == 0

    def is_section_disabled(self, section):
        """Is the section disabled in this phase?

        See get_active_and_disabled_for_section for an explanation.

        This returns a boolean.
        """
        if section == "preparation":
            return self.phase == ""
        if section == "involve":
            return self.phase == "" or (
                self.phase == "preparation" and self.is_new_session
            )
        if section in ("identification", "actionplan"):
            return (
                self.phase in ("", "preparation") and self.is_new_session
            ) or not self.can_inspect_session
        if section in ("consultancy", "report", "training", "status", "reminder"):
            # These menu items should be active even if user cannot edit or inspect.
            return self.phase in ("", "preparation") and self.is_new_session

        # Default to not disable unknown sections.
        return False

    def get_active_and_disabled_for_section(self, section):
        """Is the section active or disabled in this phase?

        The navigation tree has sections: preparation, identification, etc.
        The phase (self.phase) is the current section you are viewing.
        The section argument that is passed in, is a section in the navigation
        tree.  This method answers two questions:

        * Is the section the currently active phase?
        * Should the section be disabled?

        What 'disabled' means is up to the frontend, but the expectation is
        that a disabled section is shown greyed out, and not clickable.

        This returns a tuple with two booleans: active and disabled.
        """
        active = section == self.phase
        disabled = self.is_section_disabled(section)
        return active, disabled

    @property
    @memoize
    def survey_tree_data(self):
        if self.anonymous:
            return []

        url = self.traversed_session.absolute_url()
        integrated_action_plan = self.integrated_action_plan

        data = []

        active, disabled = self.get_active_and_disabled_for_section("preparation")
        data.append(
            {
                "active": active,
                "disabled": disabled,
                "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                "id": "step-1",
                "name": "preparation",
                "href": f"{url}/@@start#content",
                "title": api.portal.translate(
                    _("label_preparation", default="Preparation")
                ),
                "has_tree": False,
            }
        )

        if self.use_involve_phase:
            active, disabled = self.get_active_and_disabled_for_section("involve")
            data.append(
                {
                    "active": active,
                    "disabled": disabled,
                    "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                    "id": "step-involve",
                    "name": "involve",
                    "href": f"{url}/@@involve#content",
                    "title": api.portal.translate(
                        _("label_involve", default="Involve")
                    ),
                    "has_tree": False,
                }
            )

        active, disabled = self.get_active_and_disabled_for_section("identification")
        title = ""
        if integrated_action_plan:
            title = api.portal.translate(_("label_assessment", "Assessment"))
        else:
            t1 = api.portal.translate(_("label_identification", "Identification"))
            t2 = api.portal.translate(_("label_evaluation", "Evaluation"))
            title = f"{t1} + {t2}"
        data.append(
            {
                "active": active,
                "disabled": disabled,
                "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                "id": "step-2",
                "name": "identification",
                "href": f"{url}/@@identification#content",
                "title": title,
                "has_tree": active,
            }
        )

        if self.use_action_plan_phase:
            # The tree for the action section uses the same structure as the
            # identification tree, with the only differences that only risks
            # and their parent modules are shown and that the entire tree is
            # expanded.
            active, disabled = self.get_active_and_disabled_for_section("actionplan")
            data.append(
                {
                    "active": active,
                    "disabled": disabled,
                    "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                    "id": "step-4",
                    "name": "action-plan",
                    "href": f"{url}/@@actionplan#content",
                    "title": api.portal.translate(
                        _("label_action_plan", default="Action Plan")
                    ),
                    "has_tree": active,
                }
            )

        if self.use_consultancy_phase:
            active, disabled = self.get_active_and_disabled_for_section("consultancy")
            data.append(
                {
                    "active": active,
                    "disabled": disabled,
                    "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                    "id": "step-consultancy",
                    "name": "consultancy",
                    "href": f"{url}/@@consultancy#content",
                    "title": api.portal.translate(
                        _("label_consultancy", default="Consultancy")
                    ),
                    "has_tree": False,
                }
            )

        # TODO: check if guest.
        active, disabled = self.get_active_and_disabled_for_section("report")
        data.append(
            {
                "active": active,
                "disabled": disabled,
                "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                "id": "step-5",
                "name": "report",
                "href": f"{url}/@@report#content",
                "title": api.portal.translate(_("label_report", default="Report")),
                "has_tree": False,
            }
        )

        if self.use_training_module:
            active, disabled = self.get_active_and_disabled_for_section("training")
            data.append(
                {
                    "active": active,
                    "disabled": disabled,
                    "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                    "id": "step-6",
                    "name": "training",
                    "href": f"{url}/@@training#content",
                    "title": api.portal.translate(
                        _("menu_training", default="Training")
                    ),
                    "has_tree": False,
                }
            )

        active, disabled = self.get_active_and_disabled_for_section("status")
        data.append(
            {
                "active": active,
                "disabled": disabled,
                "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                "id": "status",
                "name": "status",
                "href": f"{url}/@@status#content",
                "title": api.portal.translate(_("navigation_status", default="Status")),
                "has_tree": False,
            }
        )

        if self.use_email_reminder:
            active, disabled = self.get_active_and_disabled_for_section("reminder")
            data.append(
                {
                    "active": active,
                    "disabled": disabled,
                    "class": f'{"active" if active else ""} {"disabled" if disabled else ""}',  # noqa: E501
                    "id": "reminder",
                    "name": "involve",
                    "href": f"{url}/@@email-reminder#content",
                    "title": api.portal.translate(
                        _("navigation_email_reminder", default="Email reminder")
                    ),
                    "has_tree": False,
                }
            )

        return data

    def get_user_email(self, account=None):
        return account.email

    def get_user_fullname(self, account=None):
        return account.title

    def __call__(self):
        return self



from datetime import datetime
from euphorie.client import model
from euphorie.client.tests.utils import addAccount
from euphorie.client.tests.utils import addSurvey
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from zExceptions import Unauthorized

import json


class TestExport(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.loginAsPortalOwner()
        addSurvey(self.portal, BASIC_SURVEY)
        self.jane = addAccount("jane@example.com", password="secret")
        self.john = addAccount("john@example.com", password="secret")

        survey_session = model.SurveySession(
            title="Dummy session",
            created=datetime(2012, 4, 22, 23, 5, 12),
            modified=datetime(2012, 4, 23, 11, 50, 30),
            zodb_path="nl/ict/software-development",
            account=self.jane,
            company=model.Company(country="nl", employees="1-9", referer="other"),
        )
        module = survey_session.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        risk = module.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        model.ActionPlan(action_plan="This is the plan", risk=risk)
        model.Session.add(survey_session)
        self.logout()

    def test_export_json(self):
        country = self.portal.client.nl
        traversed_session = country.ict["software-development"].restrictedTraverse(
            "++session++1"
        )
        # Calling the view as anonymous raises an Unauthorized exception
        with self._get_view("export.json", traversed_session) as view:
            with self.assertRaises(Unauthorized):
                view()

        # Calling the view as John raises an Unauthorized exception
        with api.env.adopt_user(user=self.john):
            with self._get_view("export.json", traversed_session) as view:
                with self.assertRaises(Unauthorized):
                    view()

        # Calling the view as Jane returns a json
        with api.env.adopt_user(user=self.jane):
            with self._get_view("export.json", traversed_session) as view:
                self.assertEqual(json.loads(view())["data"]["title"], "Dummy session")

        # Calling the view as admin returns a json
        with api.env.adopt_user(username="admin"):
            with self._get_view("export.json", traversed_session) as view:
                output = json.loads(view())
                self.assertEqual(output["data"]["title"], "Dummy session")
                self.assertEqual(output["data"]["created"], "2012-04-22T23:05:12")
                self.assertSetEqual(
                    set(output["children"]), {"company", "training", "tree"}
                )



from euphorie import MessageFactory as _
from euphorie.client import utils
from euphorie.client.browser.country import SessionsView
from euphorie.client.model import get_current_account
from euphorie.client.utils import CreateEmailTo
from plone import api
from plone.memoize.view import memoize
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.MailHost.MailHost import MailHostError
from z3c.saconfig import Session
from zExceptions import Unauthorized

import logging
import smtplib


logger = logging.getLogger(__name__)


class SurveySessionsView(SessionsView):
    """Template corresponds to proto:_layout/tool.html."""

    variation_class = ""

    def set_language(self):
        utils.setLanguage(self.request, self.context, self.context.language)

    @property
    @memoize
    def sessions(self):
        """Given some sessions create a tree."""
        return self.webhelpers.get_sessions_query(context=self.context).all()

    def create_survey_session(self, title, account=None, **params):
        """Create a new survey session.

        :param title: title for the new survey session.
        :type title: unicode
        :rtype: :py:class:`cls.survey_session_model` instance
        """
        if account is None:
            account = get_current_account()

        session = Session()
        sector = self.context.aq_parent
        country = sector.aq_parent
        zodb_path = f"{country.id}/{sector.id}/{self.context.id}"
        survey_session = self.survey_session_model(
            title=title,
            zodb_path=zodb_path,
            account_id=account.id,
            group_id=account.group_id,
        )
        for key in params:
            setattr(survey_session, key, params[key])
        session.add(survey_session)
        session.refresh(account)
        session.flush()  # flush so we get a session id
        return survey_session

    def __call__(self):
        if not self.account:
            raise Unauthorized()
        self.set_language()
        return self.index()


class SurveySessionsViewAnon(SurveySessionsView):
    def __call__(self):
        self.set_language()
        return self.index()


class DefaultIntroductionView(BrowserView):
    """Browser view that displays the default introduction text for a Suvey.

    It is used when the Survey does not define its own introduction
    """

    pass


class EmailReminder(BrowserView):
    """View that allows sending a reminder email with a link to the session"""

    variation_class = "variation-risk-assessment"
    email_template = ViewPageTemplateFile("templates/email_reminder_body.pt")

    def send_email(self):
        start_url = f"{self.context.absolute_url()}/@@start"
        body = self.email_template(
            tool_name=self.context.aq_parent.title, url=start_url
        )
        subject = api.portal.translate(_("OiRA tool reminder"))
        account = get_current_account()
        email_from_name = api.portal.get_registry_record("plone.email_from_name")
        email_from_address = api.portal.get_registry_record("plone.email_from_address")
        mail = CreateEmailTo(
            email_from_name,
            email_from_address,
            account.email,
            subject,
            body,
        )
        try:
            api.portal.send_email(
                body=mail,
                subject=subject,
                recipient=account.email,
                immediate=True,
            )
            logger.info(
                "Sent email reminder to %s",
                account.email,
            )
        except MailHostError as e:
            return self.log_error(
                "MailHost error sending email reminder to %s: %s", account.email, e
            )
        except smtplib.SMTPException as e:
            return self.log_error(
                "smtplib error sending email reminder to %s: %s", account.email, e
            )
        except OSError as e:
            return self.log_error(
                "Socket error sending email reminder to %s: %s", account.email, e[1]
            )
        api.portal.show_message("Email reminder has been sent", type="info")
        return self.request.response.redirect(start_url)

    def __call__(self):
        if self.request.method == "POST":
            self.send_email()
        return super().__call__()



"""Remove the completion_percentage column.

Revision ID: 20220705145832
Revises: 20220620103222
Create Date: 2022-07-05 14:55:08.235238
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20220705145832"
down_revision = "20220620103222"
branch_labels = None
depends_on = None


def upgrade():
    if has_column("session", "completion_percentage"):
        op.drop_column("session", "completion_percentage")


def downgrade():
    op.add_column(
        "session",
        sa.Column(
            "completion_percentage", sa.INTEGER(), autoincrement=False, nullable=True
        ),
    )



from base64 import encodebytes
from datetime import date
from datetime import datetime
from euphorie.client.model import ActionPlan
from euphorie.client.model import BaseObject
from euphorie.client.model import Company
from euphorie.client.model import Risk
from euphorie.client.model import Session
from euphorie.client.model import SurveySession
from euphorie.client.model import SurveyTreeItem
from euphorie.client.model import Training
from sqlalchemy.orm.query import Query
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface

import json


class ISA2DictAdapter(Interface):
    def get_children():
        """Return a dictionary containing whatever you like to be defined as
        children of this table."""

    def __call__():
        """Return a dictionary representation of this object."""


@implementer(ISA2DictAdapter)
@adapter(BaseObject)
class SA2DictAdapter:
    def __init__(self, context):
        self.context = context

    def get_children(self):
        return {}

    @property
    def columns(self):
        """Return a list of column names for this table."""
        return [column.key for column in self.context.__table__.columns]

    def __call__(self):
        table = self.context.__table__
        data = {}
        for column in self.columns:
            try:
                data[column] = getattr(self.context, column)
            except AttributeError:
                # Happens when using polymorphic tables:
                # the object might no have all the attributes set
                pass

        return {
            "table": table.name,
            "data": data,
            "children": self.get_children(),
        }


@adapter(SurveyTreeItem)
class SurveyTreeItem2DictAdapter(SA2DictAdapter):
    @property
    def columns(self):
        """This adapter is used for all the tables that inherit from
        SurveyTreeItem, namely Risk and Module.

        For those objects we want to return the columns from their own
        table plus the one of the SurveyTreeItem table.
        """
        columns = super().columns
        new_columns = [
            column.key
            for column in SurveyTreeItem.__table__.columns
            if column.key not in columns
        ]
        return columns + new_columns


@adapter(Risk)
class Risk2DictAdapter(SurveyTreeItem2DictAdapter):
    def get_children(self):
        """Overwrite the default get_children method to include the action
        plans."""
        children = super().get_children()
        children[ActionPlan.__table__.name] = (
            Session.query(ActionPlan)
            .filter(ActionPlan.risk_id == self.context.id)
            .order_by(ActionPlan.id)
        )
        return children


@adapter(SurveySession)
class SurveySession2DictAdapter(SA2DictAdapter):
    def get_children(self):
        """Overwrite the default get_children method to include:

        - the company
        - survey tree items (modules and risks)
        - training
        """
        return {
            klass.__table__.name: Session.query(klass)
            .filter(klass.session_id == self.context.id)
            .order_by(klass.id)
            for klass in [Company, SurveyTreeItem, Training]
        }


class SAJsonEncoder(json.JSONEncoder):
    __children_mapping__ = {
        "session": [Company, SurveyTreeItem, Training],
        "risk": [ActionPlan],
    }

    def default(self, obj):
        if isinstance(obj, BaseObject):
            adapter = ISA2DictAdapter(obj)
            return adapter()
        elif isinstance(obj, (datetime, date)):
            return obj.isoformat()
        elif isinstance(obj, Query):
            return obj.all()
        elif isinstance(obj, bytes):
            return encodebytes(obj).decode("ascii")
        return super().default(obj)



from Acquisition import aq_inner
from euphorie.client import model
from euphorie.client import utils
from euphorie.client.interfaces import CustomRisksModifiedEvent
from euphorie.client.navigation import FindNextQuestion
from euphorie.client.navigation import FindPreviousQuestion
from euphorie.client.navigation import getTreeData
from euphorie.content.interfaces import ICustomRisksModule
from euphorie.content.profilequestion import IProfileQuestion
from logging import getLogger
from plone import api
from plone.memoize.view import memoize
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.event import notify


logger = getLogger(__name__)


class IdentificationView(BrowserView):
    """The introduction page for a module."""

    variation_class = "variation-risk-assessment"
    phase = "identification"
    question_filter = None
    template = ViewPageTemplateFile("templates/module_identification.pt")

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def survey(self):
        """This is the survey dexterity object."""
        return self.webhelpers._survey

    @property
    def tree(self):
        return getTreeData(self.request, self.context)

    @property
    @memoize
    def next_question(self):
        """Try to understand what the next question will be."""
        return FindNextQuestion(
            self.context,
            dbsession=self.context.aq_parent.session,
            filter=self.question_filter,
        )

    @property
    @memoize
    def previous_question(self):
        """Try to understand what the previous question will be."""
        return FindPreviousQuestion(
            self.context,
            dbsession=self.context.aq_parent.session,
            filter=self.question_filter,
        )

    @property
    @memoize
    def next_question_url(self):
        """Return the URL to the next question."""
        if not self.next_question:
            return ""
        return "{parent_url}/{next_question_path}/@@{view}".format(
            parent_url=self.webhelpers.traversed_session.absolute_url(),
            next_question_path="/".join(self.next_question.short_path),
            view=self.__name__,
        )

    @property
    @memoize
    def previous_question_url(self):
        """Return the URL to the previous question."""
        if not self.previous_question:
            return ""
        return "{parent_url}/{next_question_path}/@@{view}".format(
            parent_url=self.webhelpers.traversed_session.absolute_url(),
            next_question_path="/".join(self.previous_question.short_path),
            view=self.__name__,
        )

    @property
    @memoize
    def next_phase_url(self):
        """Return the URL to the next question."""
        if self.integrated_action_plan:
            return "{parent_url}/@@report".format(
                parent_url=self.context.aq_parent.absolute_url()
            )
        else:
            return "{parent_url}/@@actionplan".format(
                parent_url=self.context.aq_parent.absolute_url()
            )

    @property
    @memoize
    def integrated_action_plan(self):
        return self.webhelpers.integrated_action_plan

    def __call__(self):
        # Render the page only if the user has inspection rights,
        # otherwise redirect to the start page of the session.
        if not self.webhelpers.can_inspect_session:
            return self.request.response.redirect(
                self.context.aq_parent.absolute_url() + "/@@start"
            )

        if self.webhelpers.redirectOnSurveyUpdate():
            return

        context = aq_inner(self.context)
        utils.setLanguage(self.request, self.survey, self.survey.language)

        module = self.webhelpers.traversed_session.restrictedTraverse(
            context.zodb_path.split("/")
        )
        if self.request.environ["REQUEST_METHOD"] == "POST":
            return self.save_and_continue(module)

        if IProfileQuestion.providedBy(module) and context.depth == 2:
            if self.next_question is None:
                url = self.next_phase_url
            else:
                url = self.next_question_url
            return self.request.response.redirect(url)

        self.title = module.title
        self.module = module
        number_files = 0
        for i in range(1, 5):
            number_files += getattr(self.module, f"file{i}", None) and 1 or 0
        self.has_files = number_files > 0
        self.next_is_actionplan = not self.next_question
        if ICustomRisksModule.providedBy(module):
            template = ViewPageTemplateFile(
                "templates/module_identification_custom.pt"
            ).__get__(self, "")
        else:
            template = self.template
        return template()

    def save_and_continue(self, module):
        """We received a POST request.

        Submit the form and figure out where to go next.
        """
        context = aq_inner(self.context)
        reply = self.request.form
        _next = reply.get("next", None)
        # In Safari browser we get a list
        if isinstance(_next, list):
            _next = _next.pop()
        if module.optional and self.webhelpers.can_edit_session:
            if "skip_children" in reply:
                context.skip_children = reply.get("skip_children")
                context.postponed = False
            else:
                context.postponed = True
            self.context.session.touch()

        if _next == "previous":
            if self.previous_question is None:
                # We ran out of questions, step back to intro page
                url = "%s/@@identification" % self.context.aq_parent.absolute_url()
                self.request.response.redirect(url)
                return
            self.request.response.redirect(self.previous_question_url)
            return

        if ICustomRisksModule.providedBy(module):
            if _next == "add_custom_risk" and self.webhelpers.can_edit_session:
                self.add_custom_risk()
                notify(CustomRisksModifiedEvent(self.context))
                risk_id = self.context.children().count()
                url = "{parent_url}/{risk_id}/@@identification".format(
                    parent_url=self.context.absolute_url(),
                    risk_id=risk_id,
                )
                return self.request.response.redirect(url)
            # We ran out of questions, proceed to the action plan
            return self.request.response.redirect(self.next_phase_url)
        if self.next_question is None:
            # We ran out of questions, proceed to the action plan
            return self.request.response.redirect(self.next_phase_url)

        self.request.response.redirect(self.next_question_url)

    def add_custom_risk(self):
        sql_risks = self.context.children()
        if sql_risks.count():
            counter_id = (
                max([int(risk.zodb_path.split("/")[-1]) for risk in sql_risks.all()])
                + 1
            )
        else:
            counter_id = 1

        # Add a new risk
        risk = model.Risk(
            comment="",
            priority=None,
            risk_id=None,
            risk_type="risk",
            skip_evaluation=True,
            title="",
            identification=None,
            training_notes="",
            custom_description="",
        )
        risk.is_custom_risk = True
        risk.skip_children = False
        risk.postponed = False
        risk.has_description = None
        risk.zodb_path = "/".join(
            [self.context.zodb_path] + ["%d" % counter_id]
        )  # There's a constraint for unique zodb_path per session
        risk.profile_index = 0  # XXX: not sure what this is for
        self.context.addChild(risk)


class ActionPlanView(BrowserView):
    """The introduction page for an :obj:`euphorie.content.module` in an action
    plan."""

    variation_class = "variation-risk-assessment"
    phase = "actionplan"
    question_filter = model.ACTION_PLAN_FILTER

    @property
    @memoize
    def webhelpers(self):
        return self.context.restrictedTraverse("webhelpers")

    @property
    @memoize
    def survey(self):
        """This is the survey dexterity object."""
        return self.webhelpers._survey

    @property
    def use_solution_direction(self):
        return utils.HasText(getattr(self.module, "solution_direction", None))

    @property
    @memoize
    def module(self):
        return self.context.aq_parent.aq_parent.restrictedTraverse(
            self.context.zodb_path.split("/")
        )

    @property
    def tree(self):
        return getTreeData(
            self.request, self.context, filter=self.question_filter, phase=self.phase
        )

    def __call__(self):
        # Render the page only if the user has edit rights,
        # otherwise redirect to the start page of the session.
        if not self.webhelpers.can_edit_session:
            return self.request.response.redirect(
                self.context.aq_parent.aq_parent.absolute_url() + "/@@start"
            )
        if self.webhelpers.redirectOnSurveyUpdate():
            return

        context = aq_inner(self.context)
        utils.setLanguage(self.request, self.survey, self.survey.language)
        if (IProfileQuestion.providedBy(self.module) and context.depth == 2) or (
            ICustomRisksModule.providedBy(self.module) and self.phase == "actionplan"
        ):
            next_question = FindNextQuestion(
                context, self.context.session, filter=self.question_filter
            )
            if next_question is None:
                if self.phase == ("identification", "evaluation"):
                    url = (
                        self.webhelpers.traversed_session.absolute_url()
                        + "/@@actionplan"
                    )
                elif self.phase == "actionplan":
                    url = self.webhelpers.traversed_session.absolute_url() + "/@@report"
            else:
                url = "{session_url}/{path}/@@actionplan".format(
                    session_url=self.webhelpers.traversed_session.absolute_url(),
                    path="/".join(next_question.short_path),
                )
            return self.request.response.redirect(url)

        self.title = self.context.title
        previous = FindPreviousQuestion(
            self.context, self.context.session, filter=self.question_filter
        )
        if previous is None:
            self.previous_url = "{}/@@{}".format(
                self.context.aq_parent.absolute_url(),
                self.phase,
            )
        else:
            self.previous_url = "{session_url}/{path}/@@{phase}".format(
                session_url=self.webhelpers.traversed_session.absolute_url(),
                path="/".join(previous.short_path),
                phase=self.phase,
            )
        next_question = FindNextQuestion(
            self.context, self.context.session, filter=self.question_filter
        )
        if next_question is None:
            self.next_url = (
                self.webhelpers.traversed_session.absolute_url() + "/@@report"
            )
        else:
            self.next_url = "{session_url}/{path}/@@{phase}".format(
                session_url=self.webhelpers.traversed_session.absolute_url(),
                path="/".join(next_question.short_path),
                phase=self.phase,
            )
        return self.index()



from collections import defaultdict
from euphorie.content import MessageFactory as _
from euphorie.content.surveygroup import ISurveyGroup
from nltk.corpus import stopwords
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import word_tokenize
from plone import api
from plone.autoform.form import AutoExtensibleForm
from plone.memoize import ram
from plone.memoize.view import memoize
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from z3c.form import button
from z3c.form import form
from zope import schema
from zope.annotation.interfaces import IAnnotations
from zope.interface import Interface


def _forever_cache_key(method, *args, **kwargs):
    return method.__name__


class ISimilarTitleSchema(Interface):
    min_similarity = schema.Float(
        title=_("Minimum similarity"),
        description=_("Minimum similarity between titles"),
        default=0.5,
        required=True,
        min=0.0,
        max=1.0,
    )

    max_similarity = schema.Float(
        title=_("Maximum similarity"),
        description=_("Maximum similarity between titles"),
        default=1.0,
        required=True,
        min=0.0,
        max=1.0,
    )


class SimilarTitles(AutoExtensibleForm, form.Form):
    """
    A form that allows you to find and display risks with similar titles.
    """

    schema = ISimilarTitleSchema
    ignoreContext = True
    form_name = _("Similar Titles OiRA Tool")
    show_form = True

    label = _("Similar Titles OiRA Tool")
    description = _("This tool allows you to find objects with similar titles.")

    @property
    def template(self):
        return self.index

    def redirect(self, target=None, message=None, message_type="info"):
        if target is None:
            target = self.context.absolute_url()
        if message is not None:
            api.portal.show_message(
                message=message, request=self.request, type=message_type
            )
        self.request.response.redirect(target)

    @property
    @memoize
    def extracted_data(self):
        """extract_data returns data and errors from the form.
        We want just the data
        """
        return self.extractData()[0]

    @property
    def min_similarity(self):
        return self.extracted_data.get("min_similarity", 0.5)

    @property
    def max_similarity(self):
        return self.extracted_data.get("max_similarity", 1.0)

    @property
    @memoize
    def tool_cache(self):
        return {}

    def get_tool_for_brain(self, brain):
        cache_key = "/".join(brain.getPath().split("/")[:6])
        cache = self.tool_cache
        value = cache.get(cache_key)
        if value is not None:
            return value
        obj = brain.getObject()
        for obj in obj.aq_chain:
            if ISurveyGroup.providedBy(obj):
                value = obj
                cache[cache_key] = value
                return value

    @ram.cache(_forever_cache_key)
    def initialize_nltk(self):
        """Initialize the NLTK library"""
        import nltk

        nltk.download("punkt")
        nltk.download("stopwords")
        nltk.download("wordnet")

    @property
    @ram.cache(_forever_cache_key)
    def published_surveys_paths(self):
        """Return a list of published survey paths"""
        brains = api.content.find(portal_type="euphorie.surveygroup")
        paths = []
        for brain in brains:
            obj = brain.getObject()
            published = getattr(obj, "published", "")
            if published:
                paths.append(f"{brain.getPath()}/{published}")
        return paths

    @property
    @memoize
    def similar_brains(self):
        self.initialize_nltk()
        brains = api.content.find(
            portal_type="euphorie.risk",
            sort_on="sortable_title",
            path=self.published_surveys_paths,
        )

        def preprocess_text(line):
            # Tokenize the text into individual words
            tokens = word_tokenize(line.lower())

            # Remove stop words
            stop_words = set(stopwords.words("german"))
            filtered_tokens = [word for word in tokens if word not in stop_words]

            # Lemmatize the words
            lemmatizer = WordNetLemmatizer()
            lemmatized_tokens = [lemmatizer.lemmatize(word) for word in filtered_tokens]

            # Join the tokens back into a single string
            preprocessed_text = " ".join(lemmatized_tokens)

            return preprocessed_text

        lines = [brain.Title for brain in brains]
        preprocessed_lines = [preprocess_text(line) for line in lines]
        # Calculate TF-IDF vectors https://en.wikipedia.org/wiki/Tf%E2%80%93idf
        vectorizer = TfidfVectorizer()
        tfidf_matrix = vectorizer.fit_transform(preprocessed_lines)

        # Calculate cosine similarity between all pairs of lines
        similarity_matrix = cosine_similarity(tfidf_matrix)

        # Group the lines that have a similarity score between
        # self.min_similarity and self.max_similarity
        similar_brains = defaultdict(list)
        min_similarity = self.min_similarity
        max_similarity = self.max_similarity
        for idx_x, brain_x in enumerate(brains):
            for idx_y, brain_y in enumerate(brains):
                # skip pairs on the diagonal, they have obviously a similarity of 1
                if idx_x != idx_y:
                    similarity = similarity_matrix[idx_x, idx_y]
                    if min_similarity <= similarity <= max_similarity:
                        similar_brains[brain_x].append((brain_y, round(similarity, 2)))
        return similar_brains

    @button.buttonAndHandler(_("Search"))
    def handle_search(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return
        self.results_html = api.content.get_view(
            context=self.context, request=self.request, name="similar-titles-results"
        )()

    @button.buttonAndHandler(_("Show stored result"))
    def handle_show_stored_result(self, action):
        self.redirect(target=f"{self.context.absolute_url()}/@@similar-titles-stored")

    @button.buttonAndHandler(_("Cancel"), name="cancel")
    def handle_cancel(self, action):
        self.redirect(message=_("Operation cancelled."))


class SimilarTitlesResults(SimilarTitles):
    @button.buttonAndHandler(_("Search"))
    def handle_search(self, action):
        data, errors = self.extractData()
        if errors:
            self.status = self.formErrorsMessage
            return


class SimilarTitlesStored(SimilarTitles):
    show_form = False

    @property
    def results_html(self):
        annotations = IAnnotations(api.portal.get())
        return annotations.get("euphorie.content.similar_titles_html", "")

    @results_html.setter
    def results_html(self, value):
        annotations = IAnnotations(api.portal.get())
        annotations["euphorie.content.similar_titles_html"] = value



"""Upgrade consultancy table

Revision ID: 20230502091628
Revises: 20230421141443
Create Date: 2023-05-02 09:24:29.093386

"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "20230502091628"
down_revision = "20230421141443"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("consultancy", "status"):
        op.add_column(
            "consultancy", sa.Column("status", sa.Unicode(length=255), nullable=True)
        )


def downgrade():
    if has_column("consultancy", "status"):
        op.drop_column("consultancy", "status")



from .survey import SurveyBase
from euphorie.content import MessageFactory as _
from plone import api
from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.browser.edit import DefaultEditForm
from plone.memoize.instance import memoize
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile


class ProfileQuestionView(SurveyBase):
    """View name: @@nuplone-view."""


class AddForm(DefaultAddForm):
    """View name: euphorie.profilequestion."""

    template = ViewPageTemplateFile("templates/profilequestion_add.pt")

    @property
    def label(self):
        return _("Add Profile question")


class AddView(DefaultAddView):
    form = AddForm


class EditForm(DefaultEditForm):
    template = ViewPageTemplateFile("templates/profilequestion_edit.pt")

    @property
    def label(self):
        return _("Edit Profile question")

    @property
    @memoize
    def portal_transforms(self):
        return api.portal.get_tool("portal_transforms")

    def get_safe_html(self, text):
        if not text:
            return ""
        data = self.portal_transforms.convertTo(
            "text/x-html-safe", text, mimetype="text/html"
        )
        return data.getData()

    def updateWidgets(self):
        super().updateWidgets()
        for fname in ("description",):
            value = self.widgets[fname].value or ""
            safe_value = self.get_safe_html(value)
            if value != safe_value:
                self.widgets[fname].value = safe_value



from .. import MessageFactory as _
from z3c.form.browser.password import PasswordWidget
from z3c.form.interfaces import IFieldWidget
from z3c.form.interfaces import IFormLayer
from z3c.form.interfaces import IPasswordWidget
from z3c.form.interfaces import IValidator
from z3c.form.validator import SimpleFieldValidator
from z3c.form.widget import FieldWidget
from zope.component import adapter
from zope.interface import implementer
from zope.interface import implementer_only
from zope.interface import Interface
from zope.schema.interfaces import IField
from zope.schema.interfaces import ValidationError


class PasswordComparisonError(ValidationError):
    __doc__ = _("Password doesn't compare with confirmation value")


class IPasswordWithConfirmationWidget(IPasswordWidget):
    """Password with confirmation widget."""


@implementer_only(IPasswordWithConfirmationWidget)
class PasswordWithConfirmationWidget(PasswordWidget):
    """Password with confirmation widget."""


@adapter(IField, IFormLayer)
@implementer(IFieldWidget)
def PasswordWithConfirmationFieldWidget(field, request):
    """IFieldWidget factory for WysiwygWidget."""
    return FieldWidget(field, PasswordWithConfirmationWidget(request))


@adapter(Interface, Interface, Interface, Interface, IPasswordWithConfirmationWidget)
@implementer(IValidator)
class PasswordWithConfirmationValidator(SimpleFieldValidator):
    def validate(self, value):
        """Check that the password value is equl to the confirmation value.

        This has been inspired by p01.widget.password
        """
        if value and self.request.get(self.widget.name + ".confirm") != value:
            raise PasswordComparisonError()
        return super(self.__class__, self).validate(value)



"""Empty message.

Revision ID: 29
Revises: 28
Create Date: 2020-12-02 09:32:52.283455
"""

from alembic import op
from euphorie.deployment.upgrade.utils import has_column

import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "29"
down_revision = "28"
branch_labels = None
depends_on = None


def upgrade():
    if not has_column("group", "active"):
        op.add_column("group", sa.Column("active", sa.Boolean(), nullable=True))
        op.execute("""UPDATE "group" set active=True""")


def downgrade():
    op.drop_column("group", "active")



from Acquisition import aq_base
from ftw.upgrade import UpgradeStep
from plone import api
from plone.uuid.interfaces import ATTRIBUTE_NAME
from plone.uuid.interfaces import IUUIDGenerator
from zope.component import getUtility

import logging


logger = logging.getLogger(__name__)


class EnsureUidIsSet(UpgradeStep):
    """Ensure that all current content has a UID."""

    def __call__(self):
        """Upgrade step to set a UID if content does not have it yet."""
        logger.info("Ensuring all items have a UID.")
        catalog = api.portal.get_tool("portal_catalog")
        generator = getUtility(IUUIDGenerator)
        missing = 0
        for brain in catalog.getAllBrains():
            if brain.UID is not None:
                continue
            obj = brain.getObject()
            if getattr(aq_base(obj), ATTRIBUTE_NAME, None):
                obj.reindexObject(idxs=["UID"])
                continue
            missing += 1
            uuid = generator()
            setattr(obj, ATTRIBUTE_NAME, uuid)
            obj.reindexObject(idxs=["UID"])
            if missing % 1000 == 0:
                logger.info(
                    "Progress: Have set a UID on %d items that were missing it.",
                    missing,
                )
        logger.info("Have set a UID on %d items that were missing it.", missing)



from plone.protect.authenticator import check
from plone.protect.auto import ProtectTransform
from zExceptions import Forbidden
from zope.globalrequest import getRequest
from zope.interface import Interface
from zope.sqlalchemy.datamanager import SessionDataManager

import logging
import os


logger = logging.getLogger(__name__)
SAFE_WRITE_KEY = "euphorie.protect.safe_sql_ids"
_marker = object()

# By default we disable CSRF protection for SQL writes, to avoid breaking
# existing code and tests.  Set the environment variable
# EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL=1 to enable CSRF protection
# for SQL writes.
EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL = os.getenv(
    "EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL", 0
)
try:
    from plone.base.utils import is_truthy

    EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL = is_truthy(
        EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL
    )
except ImportError:
    # BBB for Plone 6.1.1 and earlier
    EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL = bool(
        int(EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL)
    )


def _get_obj_key(obj):
    klass = obj.__class__
    try:
        obj_id = obj.id
    except AttributeError:
        # For example Group has group_id.
        # TODO There may be nicer ways to get the primary key.
        # Maybe somehow do that once for each table at startup,
        # or keep a small list in memory that we fill as needed.
        # Also, there could be multiple primary keys.
        primary_key = klass.__table__.primary_key.columns.keys()[0]
        obj_id = getattr(obj, primary_key)
    return (klass.__name__, obj_id)


def safeSQLWrite(obj, request=None):
    """Mark SQL object as safe to write, even in GET request.

    This is our SQLAlchemy variant of plone.protect.utils.safeWrite.

    Maybe we only need this on GET requests.  But plone.protect does it on all
    request types.  Oh, right: if you have a form that does a POST, but that form
    misses the CSRF token, you would also get a Forbidden.  So we do this on all
    requests as well.
    """
    if request is None:
        request = getRequest()
    if request is None or getattr(request, "environ", _marker) is _marker:
        # Request not found or it is a TestRequest without an environment.
        logger.debug("Could not mark object as a safe write: %s", obj)
        return
    if SAFE_WRITE_KEY not in request.environ:
        request.environ[SAFE_WRITE_KEY] = set()
    try:
        key = _get_obj_key(obj)
        if key not in request.environ[SAFE_WRITE_KEY]:
            request.environ[SAFE_WRITE_KEY].add(key)
    except AttributeError:
        logger.warning("Can't get object key to mark object as safe: %s", obj)


class IDisableCSRFProtectionForSQL(Interface):
    """Marker interface: disable auto CSRF on SQL objects in a request"""


class EuphorieProtectTransform(ProtectTransform):
    """plone.protect auto CSRF transform for Euphorie.

    The standard Plone auto CSRF protection works fine for standard Plone
    objects, but it ignores changes in SQLAlchemy items.
    This override adds such protection, but detecting changed SQLALchemy
    items, and adding them to the list of 'registered' Plone items that are
    changed by this transaction.

    These changes can be found in the session, in one of the attributes
    'dirty', 'new', 'deleted'.  For example 'dirty' may be:

        IdentitySet([<euphorie.client.model.SurveySession object at ...>])

    But when a flush of the session has happened, the changes from before the
    flush are not on the session, but are hidden in a transaction object,
    which is different:

        (Pdb) resource.tx._dirty
        <WeakKeyDictionary at 0x14aaaf390>
        (Pdb) !list(resource.tx._dirty)
        [<sqlalchemy.orm.state.InstanceState object at 0x14a7af410>]
        (Pdb) state = list(resource.tx._dirty)[0]
        (Pdb) state
        <sqlalchemy.orm.state.InstanceState object at 0x14a7af410>
        (Pdb) state.obj
        <weakref at 0x14ad29670; to 'SurveySession' at 0x14a7afcd0>
        (Pdb) state.obj()
        <euphorie.client.model.SurveySession object at 0x14a7afcd0>

    It should be fine to just add the InstanceState objects from the weak
    dictionary to the 'registered' list.  If anything is in this list on a
    GET request, we will get redirected to the confirm-action view.

    But it would be nice to add the actual SQLAlchemy items: they may have
    been marked by `plone.protect.auto.safeWrite`.

    Problem though: safeWrite fails when the object has no `_p_oid`.
    And if you would change that function, you would still need to change
    the `_check` method of the transform as it has this check:

        if getattr(obj, "_p_oid", False) in safe_oids:
            continue

    So we have created our own safeSQLWrite function and let this store the
    klass name and the id or other primary key of the sql object in a list
    in the request, similar to safe_oids list from plone.protect.
    Then we check this list of safe items below in our `_registered_objects`
    method: if an item is in the list, don't add it to the `registered` list.

    We have also defined a variant of IDisableCSRFProtection specific for
    SQLAlchemy items.  If that marker interface is set, we only return
    the `_registered_objects` of our super class, without checking for SQL
    items.  This may be needed in a few cases, at least temporarily: there may
    be some existing cases of a valid write without authenticator that worked
    so far, but that would break with our restored csrf protection.

    Note that the best way of preventing errors or warnings due to csrf
    protection, is by making sure that links and forms contain the
    authenticator.

    One more note of warning: on the client side I get a Forbidden when I
    get redirected to the confirm-action view on the Plone Site root.  A client
    side user does not have View permission there.  So we may need some changes
    there.
    """

    # Define variable in class so we can easily override it in tests.
    euphorie_enable_csrf_protection_for_sql = EUPHORIE_ENABLE_CSRF_PROTECTION_FOR_SQL

    @staticmethod
    def _get_real_objects(tx, name):
        return [state.obj() for state in getattr(tx, name, [])]

    def filter_on_safe_keys(self, registered):
        if not registered:
            return registered

        safe_keys = []
        if SAFE_WRITE_KEY in getattr(self.request, "environ", {}):
            safe_keys = self.request.environ[SAFE_WRITE_KEY]
        if not safe_keys:
            return registered

        filtered = []
        for obj in registered:
            try:
                key = _get_obj_key(obj)
            except AttributeError:
                logger.warning("Can't get object key: %s", obj)
                # If we can't get a key, it can't have been marked as safe
                # either, so we keep the object.
                filtered.append(obj)
                continue
            if key not in safe_keys:
                filtered.append(obj)
        return filtered

    def _registered_sql_objects(self):
        """Get changed SQLAlchemy objects in the current transaction.

        Here we don't care if SQL CSRF protection is enabled in general or not:
        we just return all changed SQL objects.
        But if it was explicitly disabled for this request by the marker
        interface, we return an empty list.
        """
        new_registered = []
        if IDisableCSRFProtectionForSQL.providedBy(self.request):
            return new_registered

        app = self.request.PARENTS[-1]
        for name, conn in app._p_jar.connections.items():
            if name == "temporary":
                continue
            for resource in conn.transaction_manager.get()._resources:
                if isinstance(resource, SessionDataManager):
                    # Get changed, new, and deleted items from the session.
                    for attr in ("dirty", "new", "deleted"):
                        new_registered.extend(getattr(resource.session, attr, []))
                    # Get changes that have been flushed already.
                    for attr in ("_dirty", "_new", "_deleted"):
                        new_registered.extend(self._get_real_objects(resource.tx, attr))
        return self.filter_on_safe_keys(new_registered)

    def _registered_objects(self):
        registered = super()._registered_objects()

        new_registered = self._registered_sql_objects()
        if not new_registered:
            # Return the original list.
            return registered

        if self.euphorie_enable_csrf_protection_for_sql:
            registered.extend(new_registered)
        else:
            # We still want to warn about this.  But that only makes sense
            # if the authenticator check fails.
            try:
                check(self.request, manager=self.key_manager)
            except Forbidden:
                logger.warning(
                    "CSRF protection for SQLAlchemy changes is not enabled, "
                    "allowing %d unauthenticated changes on %s request %s",
                    len(new_registered),
                    self.request.REQUEST_METHOD,
                    self.request.URL,
                )

        return registered



from euphorie.testing import EuphorieIntegrationTestCase
from plone.namedfile.file import NamedBlobImage


class PublicationTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.client = self.portal.client
        self.loginAsPortalOwner()

    def createSurvey(self):
        self.portal.sectors.nl.invokeFactory(
            "euphorie.sector", "dining", title="Fine diningÂ®"
        )
        self.sector = self.portal.sectors.nl.dining
        self.sector.invokeFactory("euphorie.surveygroup", "survey", title="Survey")
        self.surveygroup = self.sector.survey
        self.surveygroup.invokeFactory("euphorie.survey", "version1", title="Version 1")
        self.survey = self.surveygroup.version1
        return self.survey

    def testPublishEmptySurvey(self):
        self.createSurvey()
        self.surveygroup.evaluation_algorithm = "dummy"
        view = self.survey.restrictedTraverse("@@publish")
        view.publish()
        self.assertEqual(set(self.client.objectIds()), {"nl"})
        self.assertEqual(self.client.nl.objectIds(), ["dining"])
        sector = self.client.nl.dining
        self.assertEqual(sector.portal_type, "euphorie.clientsector")
        self.assertEqual(sector.Title(), "Fine diningÂ®")
        survey = sector.survey
        self.assertEqual(survey.portal_type, "euphorie.survey")
        self.assertEqual(survey.evaluation_algorithm, "dummy")
        self.assertEqual(survey.title, "Survey")
        self.assertEqual(survey.objectIds(), [])

    def testPublishSurveyWithQuestion(self):
        self.createSurvey()
        modid = self.survey.invokeFactory("euphorie.module", "module")
        module = getattr(self.survey, modid)
        qid = module.invokeFactory(
            "euphorie.risk", "risk", title="Do you offer take away?"
        )
        view = self.survey.restrictedTraverse("@@publish")
        view.publish()
        self.assertEqual(self.client.nl.objectIds(), ["dining"])
        survey = self.client.nl.dining.survey
        self.assertEqual(survey.objectIds(), [modid])
        self.assertEqual(getattr(survey, modid).objectIds(), [qid])

    def testPublishRemovesExistingObject(self):
        self.createSurvey()
        self.client.invokeFactory("euphorie.clientcountry", "nl")
        self.client.nl.invokeFactory("euphorie.clientsector", "dining")
        self.client.nl.dining.invokeFactory("euphorie.survey", "survey")
        self.client.nl.dining.survey.old_object = True
        view = self.survey.restrictedTraverse("@@publish")
        view.publish()
        self.assertEqual(getattr(self.client.nl.dining, "old_object", False), False)

    def testPublishUpdatesLogo(self):
        # Test for http://code.simplon.biz/tracker/tno-euphorie/ticket/136
        white_gif = (
            b"GIF89a\x01\x00\x01\x00\x80\x00\x00\xff\xff\xff\xff\xff\xff,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;"
        )
        black_gif = (
            b"GIF89a\x01\x00\x01\x00\x80\x01\x00\x00\x00\x00\xff\xff\xff,"
            b"\x00\x00\x00\x00\x01\x00\x01\x00\x00\x02\x02D\x01\x00;"
        )
        self.createSurvey()
        self.sector.logo = NamedBlobImage(
            data=white_gif, contentType="image/gif", filename="white.gif"
        )
        view = self.survey.restrictedTraverse("@@publish")
        view.publish()
        client_sector = self.client.nl["dining"]
        self.assertEqual(client_sector.logo.data, white_gif)
        images = client_sector.restrictedTraverse("@@images")
        white_scale = images.scale("logo", height=100, direction="up").data.data
        self.sector.logo = NamedBlobImage(
            data=black_gif, contentType="image/gif", filename="black.gif"
        )
        view.publish()
        self.assertEqual(client_sector.logo.data, black_gif)
        images = client_sector.restrictedTraverse("@@images")
        black_scale = images.scale("logo", height=100, direction="up").data.data
        self.assertNotEqual(white_scale, black_scale)


class PreviewTests(EuphorieIntegrationTestCase):
    def testPreviewGetsCorrectUrl(self):
        # Test for http://code.simplon.biz/tracker/tno-euphorie/ticket/95
        from euphorie.content.tests.utils import addSurvey
        from euphorie.content.tests.utils import createSector

        self.loginAsPortalOwner()
        sector = createSector(self.portal)
        survey = addSurvey(sector)
        view = survey.restrictedTraverse("@@preview")
        preview = view.publish()
        self.assertEqual(preview.id, "preview")
        self.assertEqual(
            preview.absolute_url(), "http://nohost/plone/client/nl/sector/preview"
        )



"""Traverse a tree item. Tree items are polymorphic entities on the main table
from euphorie.client.model.SurveyTreeItem on the column type.

We have two possible objects:
- euphorie.client.model.Module (type `module`)
- euphorie.client.model.Risk (type `risk`)
"""

from euphorie.client.adapters.session_traversal import ITraversedSurveySession
from euphorie.client.model import Session
from euphorie.client.model import SurveyTreeItem
from zExceptions import NotFound
from zope.component import adapter
from zope.publisher.interfaces.browser import IBrowserRequest
from ZPublisher.BaseRequest import DefaultPublishTraverse


@adapter(ITraversedSurveySession, IBrowserRequest)
class TraversedSessionPublishTraverser(DefaultPublishTraverse):
    """Traverser for the Survey session children."""

    def _make_path(self, pathid):
        return pathid.zfill(3)

    def publishTraverse(self, request, pathid):
        """Return a traversable SQL object."""
        query = (
            Session.query(SurveyTreeItem)
            .filter(SurveyTreeItem.session == self.context.session)
            .filter(SurveyTreeItem.path == self._make_path(pathid))
        )
        sqlobj = query.one()
        if sqlobj is None:
            raise NotFound
        return sqlobj.__of__(self.context)


@adapter(SurveyTreeItem, IBrowserRequest)
class SurveyTreeItemPublishTraverser(TraversedSessionPublishTraverser):
    """Traverser for the Survey session children."""

    def _make_path(self, pathid):
        return self.context.path + pathid.zfill(3)



from .fti import ConditionalDexterityFTI
from .fti import IConstructionFilter
from euphorie.content import MessageFactory as _
from plone import api
from plone.dexterity.content import Container
from plone.supermodel import model
from zope import schema
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


class ITrainingQuestion(model.Schema):
    """A simple schema that adds three answer fields to the content type."""

    title = schema.Text(title=_("Question"))
    right_answer = schema.Text(title=_("Right answer"))
    wrong_answer_1 = schema.Text(title=_("First wrong answer"))
    wrong_answer_2 = schema.Text(title=_("Second wrong answer"))


@implementer(ITrainingQuestion)
class TrainingQuestion(Container):
    """A Question for the training."""


@adapter(ConditionalDexterityFTI, Interface)
@implementer(IConstructionFilter)
class ConstructionFilter:
    """FTI construction filter for :py:class:`TrainingQuestion` objects. This
    filter prevents creating Training Questions if the OiRA Tool is not
    configured to provide online taining.

    This multi adapter requires the use of the conditional FTI as implemented
    by :py:class:`euphorie.content.fti.ConditionalDexterityFTI`.
    """

    def __init__(self, fti, container):
        self.fti = fti
        self.container = container

    def allowed(self):
        if not api.portal.get_registry_record(
            "euphorie.use_training_module", default=False
        ):
            return False
        return getattr(self.container, "enable_test_questions", False)



from Acquisition import aq_inner
from euphorie.content.countrymanager import ICountryManager
from euphorie.content.sector import getSurveys
from euphorie.content.sector import ISector
from euphorie.content.utils import CUSTOM_COUNTRY_NAMES
from io import StringIO
from plone.dexterity.browser.add import DefaultAddForm
from plone.dexterity.browser.add import DefaultAddView
from plone.dexterity.browser.edit import DefaultEditForm
from plone.memoize.instance import memoize
from Products.Five import BrowserView

import csv
import datetime


class CountryView(BrowserView):
    @property
    def title(self):
        names = self.request.locale.displayNames.territories
        # Hook in potential custom country names
        names.update(CUSTOM_COUNTRY_NAMES)
        return names.get(self.context.id.upper(), self.context.title)

    @property
    def sectors(self):
        sectors_dict = [
            {"id": sector.id, "title": sector.title, "url": sector.absolute_url()}
            for sector in self.context.values()
            if ISector.providedBy(sector)
        ]
        try:
            sectors_dict.sort(key=lambda s: s["title"].lower())
        except UnicodeDecodeError:
            sectors_dict.sort(key=lambda s: s["title"].lower().decode("utf-8"))
        return sectors_dict


class AddForm(DefaultAddForm):
    portal_type = "euphorie.country"

    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["country_type"].mode = "hidden"


class AddView(DefaultAddView):
    form = AddForm


class EditForm(DefaultEditForm):
    def updateWidgets(self):
        super().updateWidgets()
        self.widgets["country_type"].mode = "hidden"


class ManageUsers(BrowserView):
    @property
    @memoize
    def country(self):
        return aq_inner(self.context)

    @property
    def title(self):
        names = self.request.locale.displayNames.territories
        return names.get(self.country.id.upper(), self.country.title)

    @property
    def sectors(self):
        sectors_dict = [
            {
                "id": sector.id,
                "login": sector.login,
                "password": sector.password,
                "title": sector.title,
                "url": sector.absolute_url(),
                "locked": sector.locked,
            }
            for sector in self.country.values()
            if ISector.providedBy(sector)
        ]
        sectors_dict.sort(key=lambda s: s["title"].lower())
        return sectors_dict

    @property
    def managers(self):
        managers_dict = [
            {
                "id": manager.id,
                "login": manager.login,
                "title": manager.title,
                "url": manager.absolute_url(),
                "locked": manager.locked,
            }
            for manager in self.country.values()
            if ICountryManager.providedBy(manager)
        ]
        managers_dict.sort(key=lambda s: s["title"].lower())
        return managers_dict


class Tools(CountryView):
    @memoize
    def get_tools(self, sector_id):
        sector = self.context.get(sector_id)
        return getSurveys(sector)

    def download_csv(self):
        fieldnames = [
            "Sector",
            "Tool Title",
            "Version Title",
            "URL",
            "published",
            "obsolete",
        ]
        buffer = StringIO()
        writer = csv.DictWriter(buffer, fieldnames=fieldnames)
        writer.writeheader()
        for sector in self.sectors:
            for group in self.get_tools(sector["id"]):
                for survey in group["surveys"]:
                    writer.writerow(
                        {
                            "Sector": sector["title"],
                            "Tool Title": group["title"],
                            "Version Title": survey["title"],
                            "URL": survey["url"],
                            "published": "published" if survey["published"] else "",
                            "obsolete": "obsolete" if group["obsolete"] else "",
                        }
                    )
        csv_data = buffer.getvalue()
        buffer.close()
        response = self.request.RESPONSE
        today_iso = datetime.date.today().isoformat()
        response.setHeader(
            "Content-Disposition",
            f"attachment; filename=oira_tools_{self.context.id}_{today_iso}.csv",
        )
        response.setHeader("Content-Type", "text/csv;charset=utf-8")
        return csv_data



from euphorie.client import MessageFactory as _
from euphorie.client.model import SessionEvent
from euphorie.client.model import show_timezone
from euphorie.client.model import SurveySession
from json import JSONDecodeError
from json import loads
from logging import getLogger
from plone import api
from plone.memoize.instance import memoize
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


logger = getLogger(__name__)


class IHistoryItem(Interface):
    """"""


@implementer(IHistoryItem)
class HistoryItem:
    """An item used to build the history timeline of a session."""

    _message = ""
    css_classes = "object"
    time_attribute = ""

    def __init__(self, context):
        self.context = context

    @property
    def email(self):
        return self.context.account.login

    @property
    def raw_time(self):
        return getattr(self.context, self.time_attribute)

    @property
    def time(self):
        return self.raw_time.isoformat()

    @property
    def message(self):
        return api.portal.translate(self._message)


@adapter(SurveySession)
class SessionStarted(HistoryItem):
    _message = _("started assessment")
    time_attribute = "created"


@adapter(SurveySession)
class SessionArchived(HistoryItem):
    _message = _("archived")
    time_attribute = "archived"


@adapter(SessionEvent)
class SessionEventHistoryItem(HistoryItem):
    """An item used to build the history timeline of a session built on top of
    an SQL object."""

    time_attribute = "time"

    @property
    @memoize
    def note(self):
        """Return the additional informations for this stored event."""
        try:
            return loads(self.context.note)
        except JSONDecodeError:
            logger.warning(
                "Could not load note for session_event %r (%r)",
                self.context.id,
                self.context.note,
            )
            return {}

    @property
    def raw_time(self):
        """The dates are returned naive, localize them with the proper timezone"""
        time = getattr(self.context, self.time_attribute)
        timezone = show_timezone()
        return timezone.localize(time)

    def session_url(self):
        session = self.context.session
        return "{tool_url}/++session++{session_id}".format(
            tool_url=session.tool.absolute_url(), session_id=session.id
        )

    def email2link(self, email):
        if not email:
            return None
        return '<a href="mailto:{email}">{email}</a>'.format(email=email)


class SessionValidationRequested(SessionEventHistoryItem):
    _message = _("validation requested")


class SessionValidated(SessionEventHistoryItem):
    _message = _("validated")


class SessionInalidated(SessionEventHistoryItem):
    _message = _("invalidated")


class SessionLockSet(SessionEventHistoryItem):
    _message = _("lock set")


class SessionLockUnset(SessionEventHistoryItem):
    _message = _("lock unset")


class SessionLockRefresh(SessionEventHistoryItem):
    _message = _("lock refreshed")
